import asyncio
from unittest.mock import MagicMock, AsyncMock

import iceprod.server.config
import iceprod.server.queue


def test_queue_init():
    override = ['queue.type=test']
    cfg = iceprod.server.config.IceProdConfig(save=False, override=override)

    iceprod.server.queue.Queue(cfg=cfg)


async def test_queue_check_proxy():
    proxy_mock = MagicMock(iceprod.server.queue.SiteGlobusProxy)

    override = ['queue.type=test']
    cfg = iceprod.server.config.IceProdConfig(save=False, override=override)

    q = iceprod.server.queue.Queue(cfg=cfg)
    q.proxy = proxy_mock

    t = asyncio.create_task(q.check_proxy())
    await asyncio.sleep(0)
    t.cancel()
    await asyncio.sleep(0)

    assert proxy_mock.update_proxy.called
    assert cfg['queue']['x509proxy'] == proxy_mock.get_proxy.return_value


async def test_queue_check_proxy_duration():
    proxy_mock = MagicMock(iceprod.server.queue.SiteGlobusProxy)

    override = ['queue.type=test']
    cfg = iceprod.server.config.IceProdConfig(save=False, override=override)

    q = iceprod.server.queue.Queue(cfg=cfg)
    q.proxy = proxy_mock

    t = asyncio.create_task(q.check_proxy(duration=10*3600))
    await asyncio.sleep(0)
    t.cancel()
    await asyncio.sleep(0)

    assert proxy_mock.set_duration.called
    assert proxy_mock.set_duration.call_args == ((10,),)
    assert proxy_mock.update_proxy.called
    assert cfg['queue']['x509proxy'] == proxy_mock.get_proxy.return_value


async def test_queue_run(monkeypatch):
    proxy_mock = MagicMock(iceprod.server.queue.SiteGlobusProxy)
    monkeypatch.setattr(iceprod.server.queue, 'SiteGlobusProxy', proxy_mock)

    override = ['queue.type=test']
    cfg = iceprod.server.config.IceProdConfig(save=False, override=override)

    q = iceprod.server.queue.Queue(cfg=cfg)

    async def fn():
        await asyncio.sleep(0)
    q.grid.run = AsyncMock(side_effect=fn)

    await q.run()

    assert q.grid.run.called
    assert proxy_mock.return_value.update_proxy.called
    assert cfg['queue']['x509proxy'] == proxy_mock.return_value.get_proxy.return_value



"""
Test script for common functions
"""

from __future__ import absolute_import, division, print_function

from tests.util import unittest_reporter, glob_tests

import logging
logger = logging.getLogger('functions')

import os
import shutil
import random
import string
import subprocess
import tempfile

try:
    pass
except:
    pass

try:
    import psutil
except ImportError:
    psutil = None

import unittest
try:
    from unittest.mock import patch
except ImportError:
    from mock import patch
import requests_mock

from tornado.testing import AsyncTestCase

import iceprod.core.dataclasses
import iceprod.core.util
import iceprod.core.gridftp
import iceprod.core.functions


class functions_test(AsyncTestCase):
    def setUp(self):
        super(functions_test,self).setUp()

        self.test_dir = tempfile.mkdtemp(dir=os.getcwd())
        curdir = os.getcwd()
        os.symlink(os.path.join(curdir, 'iceprod'),
                   os.path.join(self.test_dir, 'iceprod'))
        os.chdir(self.test_dir)
        def cleanup():
            os.chdir(curdir)
            shutil.rmtree(self.test_dir)
        self.addCleanup(cleanup)

    @unittest_reporter
    def test_001_uncompress(self):
        """Test uncompressing a file"""
        for ext in ('gz','bz2','xz','lzma'):
            for i in range(10):
                # create test file
                filename = os.path.join(self.test_dir,'test_uncompress'+str(random.randint(0,100000)))
                while os.path.exists(filename):
                    filename = os.path.join(self.test_dir,'test_uncompress'+str(random.randint(0,100000)))
                with open(filename,'w') as f:
                    file_contents = ''
                    for x in range(1000):
                        file_contents += str(random.choice(string.ascii_letters))
                    f.write(file_contents)

                # compress
                outfile = iceprod.core.functions.compress(filename, ext)
                logger.info('compressed file is %s'%outfile)
                if outfile != filename+'.'+ext:
                    raise Exception('did not create correct filename')
                if not os.path.isfile(outfile):
                    raise Exception('did not create compressed file')

                # remove original file
                if os.path.exists(filename):
                    os.remove(filename)

                # uncompress
                files = iceprod.core.functions.uncompress(outfile)
                logger.info('uncompress returned %r',files)
                if not isinstance(files,str):
                    raise Exception('uncompress did not return a string of the new file name')
                if not os.path.isfile(files):
                    raise Exception('uncompress returned an invalid file name')

                # check file
                if not os.path.samefile(filename,files):
                    logger.warning('file names are different')
                with open(files, 'r') as f:
                    results = f.read(len(file_contents)*10)
                    if file_contents != results:
                        raise Exception('contents not the same')

    @unittest_reporter(name='uncompress() with tar files')
    def test_002_uncompress_tar(self):
        """Test uncompressing a file with tar"""
        for ext in ('tar.gz','tgz','tar.bz2','tbz','tar.xz','tar.lzma'):
            for i in range(10):
                filename = os.path.join(self.test_dir,'test_uncompress'+str(random.randint(0,100000)))
                while os.path.exists(filename):
                    filename = os.path.join(self.test_dir,'test_uncompress'+str(random.randint(0,100000)))
                os.mkdir(filename)
                infiles = {}
                for _ in range(10):
                    # create test file
                    fname = os.path.join(filename,str(random.randint(0,100000)))
                    with open(fname,'w') as f:
                        file_contents = ''
                        for x in range(1000):
                            file_contents += str(random.choice(string.ascii_letters))
                        f.write(file_contents)
                    infiles[fname] = file_contents

                # compress
                outfile = iceprod.core.functions.compress(filename, ext)
                logger.info('compressed file is %s'%outfile)
                if outfile != filename+'.'+ext:
                    raise Exception('did not create correct filename')
                if not os.path.isfile(outfile):
                    raise Exception('did not create compressed file')

                # remove original file
                if os.path.exists(filename):
                    shutil.rmtree(filename)

                # uncompress
                files = iceprod.core.functions.uncompress(outfile)
                logger.info('uncompress returned %r',files)
                if set(files) != set(x.replace(self.test_dir+'/','') for x in infiles):
                    raise Exception('not the same files')
                
                # check files
                for fname in infiles:
                    if not os.path.exists(fname):
                        logger.warning('file names are different')
                    with open(fname, 'r') as f:
                        results = f.read(len(file_contents)*10)
                        if infiles[fname] != results:
                            raise Exception('contents not the same')

    @unittest_reporter(name='uncompress() tar - other dir')
    def test_003_uncompress_tar(self):
        """Test uncompressing a file with tar"""
        for ext in ('tar.gz','tgz','tar.bz2','tbz','tar.xz','tar.lzma'):
            filename = os.path.join(self.test_dir,'test_uncompress'+str(random.randint(0,100000)))
            while os.path.exists(filename):
                filename = os.path.join(self.test_dir,'test_uncompress'+str(random.randint(0,100000)))
            os.mkdir(filename)
            infiles = {}
            for _ in range(10):
                # create test file
                fname = os.path.join(filename,str(random.randint(0,100000)))
                with open(fname,'w') as f:
                    file_contents = ''
                    for x in range(1000):
                        file_contents += str(random.choice(string.ascii_letters))
                    f.write(file_contents)
                infiles[fname] = file_contents

            # compress
            outfile = iceprod.core.functions.compress(filename, ext)
            logger.info('compressed file is %s'%outfile)
            if outfile != filename+'.'+ext:
                raise Exception('did not create correct filename')
            if not os.path.isfile(outfile):
                raise Exception('did not create compressed file')

            # remove original file
            if os.path.exists(filename):
                shutil.rmtree(filename)

            local_dir = os.path.join(self.test_dir, 'local')
            if not os.path.exists(local_dir):
                os.makedirs(local_dir)

            # uncompress
            files = iceprod.core.functions.uncompress(outfile, out_dir=local_dir)
            logger.info('uncompress returned %r',files)
            if set(files) != set(x.replace(self.test_dir+'/','') for x in infiles):
                raise Exception('not the same files')

            # check that files are extracted in local dir
            logger.info('local dir %r: %r', local_dir, os.listdir(local_dir))
            if files[0].split(os.path.sep)[0] not in os.listdir(local_dir):
                raise Exception('did not extract in local dir')

            # check files
            for fname in infiles:
                common = os.path.commonprefix([local_dir, fname])
                fname2 = os.path.join(local_dir, fname[len(common):])
                if not os.path.exists(fname2):
                    raise Exception('file %r does not exist'%fname2)
                with open(fname2, 'r') as f:
                    results = f.read(len(file_contents)*10)
                    if infiles[fname] != results:
                        raise Exception('contents not the same')


    @unittest_reporter
    def test_020_iscompressed(self):
        """Test the iscompressed function with various extensions"""
        for i in range(0,10):
            if not iceprod.core.functions.iscompressed('test.gz'):
                raise Exception('failed on .gz')
            if not iceprod.core.functions.iscompressed('test.tar.gz'):
                raise Exception('failed on .tar.gz')
            if not iceprod.core.functions.iscompressed('test.tgz'):
                raise Exception('failed on .tgz')
            if not iceprod.core.functions.iscompressed('test.tar.bz2'):
                raise Exception('failed on .tar.bz2')
            if not iceprod.core.functions.iscompressed('test.lzma'):
                raise Exception('failed on .lzma')
            if not iceprod.core.functions.iscompressed('test.xz'):
                raise Exception('failed on .xz')
            if not iceprod.core.functions.iscompressed('test.tar.lzma'):
                raise Exception('failed on .tar.lzma')
            if iceprod.core.functions.iscompressed('test'):
                raise Exception('failed on (no ext)')
            if iceprod.core.functions.iscompressed('test.doc'):
                raise Exception('failed on .doc')
            if iceprod.core.functions.iscompressed('test.xml'):
                raise Exception('failed on .xml')
            if iceprod.core.functions.iscompressed('test.gzhelp'):
                raise Exception('failed on .gzhelp')

    @unittest_reporter
    def test_021_istarred(self):
        """Test the istarred function with various extensions"""
        for i in range(0,10):
            if not iceprod.core.functions.istarred('test.tar.gz'):
                raise Exception('failed on .tar.gz')
            if not iceprod.core.functions.istarred('test.tgz'):
                raise Exception('failed on .tgz')
            if not iceprod.core.functions.istarred('test.tar.bz2'):
                raise Exception('failed on .tar.bz2')
            if not iceprod.core.functions.istarred('test.tar.lzma'):
                raise Exception('failed on .tar.lzma')
            if iceprod.core.functions.istarred('test'):
                raise Exception('failed on (no ext)')
            if iceprod.core.functions.istarred('test.doc'):
                raise Exception('failed on .doc')
            if iceprod.core.functions.istarred('test.xml'):
                raise Exception('failed on .xml')
            if iceprod.core.functions.istarred('test.gzhelp'):
                raise Exception('failed on .gzhelp')

    @unittest_reporter
    def test_100_md5sum(self):
        """Test the creation of md5sums of files"""
        for i in range(0,10):
            # create test file
            filename = os.path.join(self.test_dir,'test_md5sum'+str(random.randint(0,100000)))
            while os.path.exists(filename):
                filename = os.path.join(self.test_dir,'test_md5sum'+str(random.randint(0,100000)))
            with open(filename,'w') as f:
                file_contents = ''
                for x in range(0,1000):
                    file_contents += str(random.choice(string.ascii_letters))
                f.write(file_contents)

            # get md5sum from functions
            internal = iceprod.core.functions.md5sum(filename)
            out = subprocess.Popen('md5sum %s'%filename,shell=True,stdout=subprocess.PIPE).communicate()[0]
            try:
                external, file = out.decode('utf-8').split()
            except Exception:
                raise Exception('failed to get external md5sum')

            if internal != external:
                raise Exception('failed md5sum check')

            os.remove(filename)

    @unittest_reporter
    def test_101_check_md5sum(self):
        """Test the checking of md5sums of files"""
        for i in range(0,10):
            # create test file
            filename = os.path.join(self.test_dir,'test_md5sum'+str(random.randint(0,100000)))
            while os.path.exists(filename):
                filename = os.path.join(self.test_dir,'test_md5sum'+str(random.randint(0,100000)))
            with open(filename,'w') as f:
                file_contents = ''
                for x in range(0,1000):
                    file_contents += str(random.choice(string.ascii_letters))
                f.write(file_contents)

            # get md5sum from functions
            internal = iceprod.core.functions.md5sum(filename)

            # check md5sum
            if not iceprod.core.functions.check_md5sum(filename,internal):
                raise Exception('md5sum as str failed')

            os.remove(filename)

        for i in range(0,10):
            # create test file
            filename = os.path.join(self.test_dir,'test_md5sum'+str(random.randint(0,100000)))
            while os.path.exists(filename):
                filename = os.path.join(self.test_dir,'test_md5sum'+str(random.randint(0,100000)))
            with open(filename,'w') as f:
                file_contents = ''
                for x in range(0,1000):
                    file_contents += str(random.choice(string.ascii_letters))
                f.write(file_contents)

            # get md5sum from functions
            if subprocess.call('md5sum %s > %s.md5sum'%(filename,filename),shell=True):
                raise Exception('failed to generate md5sum')

            # check md5sum
            if not iceprod.core.functions.check_md5sum(filename,filename+'.md5sum'):
                raise Exception('md5sum as list failed')

            os.remove(filename)
            os.remove(filename+'.md5sum')

    @unittest_reporter
    def test_102_sha1sum(self):
        """Test the creation of sha1sums of files"""
        for i in range(0,10):
            # create test file
            filename = os.path.join(self.test_dir,'test_sha1sum'+str(random.randint(0,100000)))
            while os.path.exists(filename):
                filename = os.path.join(self.test_dir,'test_sha1sum'+str(random.randint(0,100000)))
            with open(filename,'w') as f:
                file_contents = ''
                for x in range(0,1000):
                    file_contents += str(random.choice(string.ascii_letters))
                f.write(file_contents)

            # get sha1sum from functions
            internal = iceprod.core.functions.sha1sum(filename)
            out = subprocess.Popen('sha1sum %s'%filename,shell=True,stdout=subprocess.PIPE).communicate()[0]
            try:
                external, file = out.decode('utf-8').split()
            except Exception:
                raise Exception('failed to get external sha1sum')

            if internal != external:
                raise Exception('failed sha1sum check')

            os.remove(filename)

    @unittest_reporter
    def test_103_check_sha1sum(self):
        """Test the checking of sha1sums of files"""
        for i in range(0,10):
            # create test file
            filename = os.path.join(self.test_dir,'test_sha1sum'+str(random.randint(0,100000)))
            while os.path.exists(filename):
                filename = os.path.join(self.test_dir,'test_sha1sum'+str(random.randint(0,100000)))
            with open(filename,'w') as f:
                file_contents = ''
                for x in range(0,1000):
                    file_contents += str(random.choice(string.ascii_letters))
                f.write(file_contents)

            # get sha1sum from functions
            internal = iceprod.core.functions.sha1sum(filename)

            # check sha1sum
            if not iceprod.core.functions.check_sha1sum(filename,internal):
                raise Exception('sha1sum as str failed')

            os.remove(filename)

        for i in range(0,10):
            # create test file
            filename = os.path.join(self.test_dir,'test_sha1sum'+str(random.randint(0,100000)))
            with open(filename,'w') as f:
                file_contents = ''
                for x in range(0,1000):
                    file_contents += str(random.choice(string.ascii_letters))
                f.write(file_contents)

            # get sha1sum from functions
            if subprocess.call('sha1sum %s > %s.sha1sum'%(filename,filename),shell=True):
                raise Exception('failed to generate sha1sum')

            # check sha1sum
            if not iceprod.core.functions.check_sha1sum(filename,filename+'.sha1sum'):
                raise Exception('sha1sum as list failed')

            os.remove(filename)
            os.remove(filename+'.sha1sum')

    @unittest_reporter
    def test_104_sha256sum(self):
        """Test the creation of sha256sums of files"""
        for i in range(0,10):
            # create test file
            filename = os.path.join(self.test_dir,'test_sha256sum'+str(random.randint(0,100000)))
            while os.path.exists(filename):
                filename = os.path.join(self.test_dir,'test_sha256sum'+str(random.randint(0,100000)))
            with open(filename,'w') as f:
                file_contents = ''
                for x in range(0,1000):
                    file_contents += str(random.choice(string.ascii_letters))
                f.write(file_contents)

            # get sha256sum from functions
            internal = iceprod.core.functions.sha256sum(filename)
            out = subprocess.Popen('sha256sum %s'%filename,shell=True,stdout=subprocess.PIPE).communicate()[0]
            try:
                external, file = out.decode('utf-8').split()
            except Exception:
                raise Exception('failed to get external sha256sum')

            if internal != external:
                raise Exception('failed sha256sum check')

            os.remove(filename)

    @unittest_reporter
    def test_105_check_sha256sum(self):
        """Test the checking of sha256sums of files"""
        for i in range(0,10):
            # create test file
            filename = os.path.join(self.test_dir,'test_sha256sum'+str(random.randint(0,100000)))
            while os.path.exists(filename):
                filename = os.path.join(self.test_dir,'test_sha256sum'+str(random.randint(0,100000)))
            with open(filename,'w') as f:
                file_contents = ''
                for x in range(0,1000):
                    file_contents += str(random.choice(string.ascii_letters))
                f.write(file_contents)

            # get sha256sum from functions
            internal = iceprod.core.functions.sha256sum(filename)

            # check sha256sum
            if not iceprod.core.functions.check_sha256sum(filename,internal):
                raise Exception('sha256sum as str failed')

            os.remove(filename)

        for i in range(0,10):
            # create test file
            filename = os.path.join(self.test_dir,'test_sha256sum'+str(random.randint(0,100000)))
            while os.path.exists(filename):
                filename = os.path.join(self.test_dir,'test_sha256sum'+str(random.randint(0,100000)))
            with open(filename,'w') as f:
                file_contents = ''
                for x in range(0,1000):
                    file_contents += str(random.choice(string.ascii_letters))
                f.write(file_contents)

            # get sha256sum from functions
            if subprocess.call('sha256sum %s > %s.sha256sum'%(filename,filename),shell=True):
                raise Exception('failed to generate sha256sum')

            # check sha256sum
            if not iceprod.core.functions.check_sha256sum(filename,filename+'.sha256sum'):
                raise Exception('sha256sum as list failed')

            os.remove(filename)
            os.remove(filename+'.sha256sum')

    @unittest_reporter
    def test_106_sha512sum(self):
        """Test the creation of sha512sums of files"""
        for i in range(0,10):
            # create test file
            filename = os.path.join(self.test_dir,'test_sha512sum'+str(random.randint(0,100000)))
            while os.path.exists(filename):
                filename = os.path.join(self.test_dir,'test_sha512sum'+str(random.randint(0,100000)))
            with open(filename,'w') as f:
                file_contents = ''
                for x in range(0,1000):
                    file_contents += random.choice(string.ascii_letters)
                f.write(file_contents)

            # get sha512sum from functions
            internal = iceprod.core.functions.sha512sum(filename)
            out = subprocess.Popen('sha512sum %s'%filename,shell=True,stdout=subprocess.PIPE).communicate()[0]
            try:
                external, file = out.decode('utf-8').split()
            except Exception:
                raise Exception('failed to get external sha512sum')

            self.assertEqual(internal, external)

            os.remove(filename)

    @unittest_reporter
    def test_107_check_sha512sum(self):
        """Test the checking of sha512sums of files"""
        for i in range(0,10):
            # create test file
            filename = os.path.join(self.test_dir,'test_sha512sum'+str(random.randint(0,100000)))
            while os.path.exists(filename):
                filename = os.path.join(self.test_dir,'test_sha512sum'+str(random.randint(0,100000)))
            with open(filename,'w') as f:
                file_contents = ''
                for x in range(0,1000):
                    file_contents += str(random.choice(string.ascii_letters))
                f.write(file_contents)

            # get sha512sum from functions
            internal = iceprod.core.functions.sha512sum(filename)

            # check sha512sum
            if not iceprod.core.functions.check_sha512sum(filename,internal):
                raise Exception('sha512sum as str failed')

            os.remove(filename)

        for i in range(0,10):
            # create test file
            filename = os.path.join(self.test_dir,'test_sha512sum'+str(random.randint(0,100000)))
            while os.path.exists(filename):
                filename = os.path.join(self.test_dir,'test_sha512sum'+str(random.randint(0,100000)))
            with open(filename,'w') as f:
                file_contents = ''
                for x in range(0,1000):
                    file_contents += str(random.choice(string.ascii_letters))
                f.write(file_contents)

            # get sha512sum from functions
            if subprocess.call('sha512sum %s > %s.sha512sum'%(filename,filename),shell=True):
                raise Exception('failed to generate sha512sum')

            # check sha512sum
            if not iceprod.core.functions.check_sha512sum(filename,filename+'.sha512sum'):
                raise Exception('sha512sum as list failed')

            os.remove(filename)
            os.remove(filename+'.sha512sum')

    @unittest_reporter
    def test_200_removedirs(self):
        """Test removing files and directories"""
        for i in range(0,10):
            # create test file
            filename = os.path.join(self.test_dir,'test_removedirs'+str(random.randint(0,100000)))
            while os.path.exists(filename):
                filename = os.path.join(self.test_dir,'test_removedirs'+str(random.randint(0,100000)))
            with open(filename,'w') as f:
                file_contents = ''
                for x in range(0,1000):
                    file_contents += str(random.choice(string.ascii_letters))
                f.write(file_contents)

            # remove file
            iceprod.core.functions.removedirs(filename)

            # check file
            if os.path.exists(filename):
                raise Exception('removedirs failed to remove %s')%filename

        for i in range(0,10):
            # create test directory
            dir = os.path.join(self.test_dir,str(random.randint(0,100000)))
            while os.path.exists(dir):
                dir = os.path.join(self.test_dir,str(random.randint(0,100000)))
            os.mkdir(dir)

            # create a file in the directory
            filename = os.path.join(dir,'test_removedirs'+str(random.randint(0,100000)))
            with open(filename,'w') as f:
                file_contents = ''
                for x in range(0,1000):
                    file_contents += str(random.choice(string.ascii_letters))
                f.write(file_contents)

            # remove directory
            iceprod.core.functions.removedirs(dir)

            # check file
            if os.path.exists(dir):
                raise Exception('removedirs failed to remove %s')%dir

    @unittest_reporter
    def test_201_copy(self):
        """Test copying files and directories"""
        for i in range(0,10):
            # create test file
            filename = os.path.join(self.test_dir,'test_copy'+str(random.randint(0,100000)))
            while os.path.exists(filename):
                filename = os.path.join(self.test_dir,'test_copy'+str(random.randint(0,100000)))
            filename2 = os.path.join(self.test_dir,'test_copy'+str(random.randint(0,100000)))
            while os.path.exists(filename2) or filename == filename2:
                filename2 = os.path.join(self.test_dir,'test_copy'+str(random.randint(0,100000)))
            with open(filename,'w') as f:
                file_contents = ''
                for x in range(0,1000):
                    file_contents += str(random.choice(string.ascii_letters))
                f.write(file_contents)

            # copy file
            iceprod.core.functions.copy(filename,filename2)

            # check file
            if not os.path.isfile(filename2):
                raise Exception('copy failed to copy %s to %s')%(filename,filename2)
            with open(filename2,'r') as f:
                results = f.read(len(file_contents)*10)
                if file_contents != results:
                    raise Exception('contents not the same')

        for i in range(0,10):
            # create test directory
            dir = os.path.join(self.test_dir,str(random.randint(0,100000)))
            while os.path.exists(dir):
                dir = os.path.join(self.test_dir,str(random.randint(0,100000)))
            dir2 = os.path.join(self.test_dir,str(random.randint(0,100000)))
            while os.path.exists(dir2) or dir == dir2:
                dir2 = os.path.join(self.test_dir,str(random.randint(0,100000)))
            os.mkdir(dir)

            # create a file in the directory
            rand = str(random.randint(0,100000))
            filename = os.path.join(dir,'test_copy'+rand)
            filename2 = os.path.join(dir2,'test_copy'+rand)
            with open(filename,'w') as f:
                file_contents = ''
                for x in range(0,1000):
                    file_contents += str(random.choice(string.ascii_letters))
                f.write(file_contents)

            # copy directory
            iceprod.core.functions.copy(dir,dir2)

            # check directory
            if not os.path.isdir(dir2):
                raise Exception('copy failed to copy %s to %s')%(dir,dir2)
            # check file
            if not os.path.isfile(filename2):
                raise Exception('copy failed to copy %s to %s')%(filename,filename2)
            with open(filename2,'r') as f:
                results = f.read(len(file_contents)*10)
                if file_contents != results:
                    raise Exception('contents not the same')

    @unittest_reporter(skip=not psutil)
    def test_300_getInterfaces(self):
        """Test the getInterfaces function"""
        # get interfaces
        ifaces = iceprod.core.functions.getInterfaces()
        if not ifaces:
            raise Exception('getInterfaces() returned None')

        loop = None
        eth = []
        for name in ifaces:
            if name == 'lo':
                loop = ifaces[name]
            else:
                eth.append(ifaces[name])

        # check that we can see the loopback interface
        if not loop:
            raise Exception('No loop interface')

        # check that we can see an ethernet interface
        if len(eth) < 1:
            raise Exception('No ethernet interfaces')
        for e in eth:
            logging.info('%s',e)

    @patch('socket.getfqdn')
    @patch('socket.gethostname')
    @unittest_reporter
    def test_301_gethostname(self, fqdn, hostname):
        fqdn.return_value = 'myhost'
        hostname.return_value = 'myhost'
        host = iceprod.core.functions.gethostname()
        self.assertEqual(host, 'myhost')

        hostname.return_value = 'myhost.test.com'
        host = iceprod.core.functions.gethostname()
        self.assertEqual(host, 'myhost.test.com')

        fqdn.return_value = 'myhost.foo.bar'
        host = iceprod.core.functions.gethostname()
        self.assertEqual(host, 'myhost.test.com')

        fqdn.return_value = 'myhost.foo.bar.baz'
        host = iceprod.core.functions.gethostname()
        self.assertEqual(host, 'myhost.foo.bar.baz')

    @unittest_reporter
    def test_302_isurl(self):
        """Test the isurl function"""
        good_urls = ['http://www.google.com',
                     'https://skua.icecube.wisc.edu:9080',
                     'http://test.com?blah=1#60',
                     'gsiftp://gridftp.icecube.wisc.edu',
                     'ftp://x2100.icecube.wisc.edu',
                     'file:/data/exp']
        bad_urls = ['slkdjf:/sldfjlksd',
                    'rpc://test']
        for i in range(0,10):
            for url in good_urls:
                if not iceprod.core.functions.isurl(url):
                    raise Exception('isurl thought %s was not a valid url'%url)
            for url in bad_urls:
                if iceprod.core.functions.isurl(url):
                    raise Exception('isurl thought %s was a valid url'%url)

    @requests_mock.mock()
    @unittest_reporter(name='download() http')
    async def test_303_download(self, http_mock):
        """Test the download function"""
        download_options = {'username': 'user',
                            'password': 'pass',}

        data = b'the data'
        md5sum = b'3d5f3303ed6ce28c2d5ac1192118f0e2'

        # download file from resources
        http_mock.get('/globus.tar.gz', content=data)
        http_mock.get('/globus.tar.gz.md5sum', content=md5sum+b' globus.tar.gz')
        await iceprod.core.functions.download('http://prod-exe.icecube.wisc.edu/globus.tar.gz',
                self.test_dir,options=download_options)
        if not os.path.isfile(os.path.join(self.test_dir,'globus.tar.gz')):
            raise Exception('downloaded file does not exist')
        self.assertTrue(http_mock.called)
        data2 = open(os.path.join(self.test_dir,'globus.tar.gz'),'rb').read()
        self.assertEqual(data2, data, msg='data not equal')

    @unittest_reporter(name='download() file')
    async def test_304_download(self):
        """Test the download function"""
        data = 'the data'
        md5sum = '3d5f3303ed6ce28c2d5ac1192118f0e2'

        # download file from local file system
        filename = os.path.join(self.test_dir,'generators.py')
        with open(filename, 'w') as f:
            f.write(data)
        await iceprod.core.functions.download(filename,
                os.path.join(self.test_dir,'generators2.py'))
        if not os.path.isfile(os.path.join(self.test_dir,'generators2.py')):
            raise Exception('local cp: copied file does not exist')
        data2 = open(os.path.join(self.test_dir,'generators2.py')).read()
        self.assertEqual(data2, data, msg='data not equal')

    @patch('iceprod.core.functions.GridFTP')
    @unittest_reporter(name='download() gridftp')
    async def test_305_download(self, gridftp):
        """Test the download function"""
        # download file from gsiftp
        data = 'the data'
        def get(url,filename=None):
            logger.info('fake get: url=%r, filename=%r', url, filename)
            if url.endswith('globus.tar.gz'):
                with open(filename,'w') as f:
                    f.write(data)
                get.url = url
            else:
                raise Exception()
        get.url = None
        gridftp.get = get

        await iceprod.core.functions.download('gsiftp://data.icecube.wisc.edu/data/sim/sim-new/downloads/globus.tar.gz',
                self.test_dir)
        if not os.path.isfile(os.path.join(self.test_dir,'globus.tar.gz')):
            raise Exception('gsiftp: downloaded file does not exist')
        data2 = open(os.path.join(self.test_dir,'globus.tar.gz')).read()
        self.assertEqual(get.url, 'gsiftp://data.icecube.wisc.edu/data/sim/sim-new/downloads/globus.tar.gz')
        self.assertEqual(data2, data, msg='data not equal')

    @requests_mock.mock()
    @unittest_reporter(name='download() http - query params')
    async def test_306_download(self, http_mock):
        """Test the download function"""
        data = b'the data'
        md5sum = b'3d5f3303ed6ce28c2d5ac1192118f0e2'

        # download file from resources
        url = 'http://prod-exe.icecube.wisc.edu/globus.tar.gz?a=1'
        http_mock.get('/globus.tar.gz?a=1', content=data)
        http_mock.get('/globus.tar.gz.md5sum', content=md5sum+b' globus.tar.gz')
        out_file = await iceprod.core.functions.download(url, self.test_dir)
        self.assertEqual(out_file, os.path.join(self.test_dir,'globus.tar.gz'))
        if not os.path.isfile(out_file):
            raise Exception('downloaded file does not exist')
        self.assertTrue(http_mock.called)
        data2 = open(out_file,'rb').read()
        self.assertEqual(data2, data, msg='data not equal')

    @unittest_reporter(name='download() errors')
    async def test_320_download(self):
        """Test the download function"""
        data = b'the data'
        md5sum = b'3d5f3303ed6ce28c2d5ac1192118f0e2'
        download_options = {}

        filename = os.path.join(self.test_dir, 'generators.py')
        out_dir = os.path.join(self.test_dir, 'output')
        os.makedirs(out_dir)
        output_file = os.path.join(out_dir, 'generators.py')
        with open(filename, 'wb') as f:
            f.write(data)
        with open(filename+'.md5sum', 'wb') as f:
            f.write(md5sum+b' generators.py')

        # bad url
        with self.assertRaises(Exception):
            await iceprod.core.functions.download(filename+'blah', out_dir)

    @requests_mock.mock()
    @unittest_reporter(name='upload() http')
    async def test_402_upload(self, http_mock):
        """Test the upload function"""
        download_options = {'username': 'user',
                            'password': 'pass',}
        data = b'the data'

        # upload file to http
        http_mock.put('/globus.tar.gz', content=b'')
        http_mock.get('/globus.tar.gz', content=data)
        filename = os.path.join(self.test_dir, 'globus.tar.gz')
        with open(filename, 'wb') as f:
            f.write(data)
        await iceprod.core.functions.upload(filename,
                'http://prod-exe.icecube.wisc.edu/globus.tar.gz',
                options=download_options)
        self.assertTrue(http_mock.called)
        req = http_mock.request_history[0]
        self.assertEqual(req.method, 'PUT', msg='not a PUT request')
        self.assertEqual(os.path.basename(req.url), 'globus.tar.gz', msg='bad upload url')

        # test bad upload
        http_mock.get('/globus.tar.gz', content=b'blah')
        with self.assertRaises(Exception):
            await iceprod.core.functions.upload(filename,
                    'http://prod-exe.icecube.wisc.edu/globus.tar.gz',
                    options=download_options)

    @requests_mock.mock()
    @unittest_reporter(name='upload() http POST')
    async def test_403_upload(self, http_mock):
        """Test the upload function"""
        download_options = {'username': 'user',
                            'password': 'pass',}
        data = b'the data'

        # upload file to http
        http_mock.put('/globus.tar.gz', content=b'', status_code=405)
        http_mock.post('/globus.tar.gz', content=b'')
        http_mock.get('/globus.tar.gz', content=data)
        filename = os.path.join(self.test_dir, 'globus.tar.gz')
        with open(filename, 'wb') as f:
            f.write(data)
        await iceprod.core.functions.upload(filename,
                'http://prod-exe.icecube.wisc.edu/globus.tar.gz',
                options=download_options)
        self.assertTrue(http_mock.called)
        req = http_mock.request_history[0]
        self.assertEqual(req.method, 'PUT', msg='not a PUT request first')
        self.assertEqual(os.path.basename(req.url), 'globus.tar.gz', msg='bad upload url')
        req = http_mock.request_history[1]
        self.assertEqual(req.method, 'POST', msg='not a POST request second')
        self.assertEqual(os.path.basename(req.url), 'globus.tar.gz', msg='bad upload url')

        # test bad upload
        http_mock.get('/globus.tar.gz', content=b'blah')
        with self.assertRaises(Exception):
            await iceprod.core.functions.upload(filename,
                    'http://prod-exe.icecube.wisc.edu/globus.tar.gz',
                    options=download_options)

    @requests_mock.mock()
    @unittest_reporter(name='upload() http s3 ETAG')
    async def test_403_upload(self, http_mock):
        """Test the upload function with ETAG"""
        download_options = {}
        data = b'the data'
        filename = os.path.join(self.test_dir, 'globus.tar.gz')
        with open(filename, 'wb') as f:
            f.write(data)

        # upload file to http
        http_mock.put('/globus.tar.gz', content=b'', headers={'ETAG': iceprod.core.functions.md5sum(filename)})
        http_mock.get('/globus.tar.gz', content=b'', status_code=403)
        await iceprod.core.functions.upload(filename,
                'http://prod-exe.icecube.wisc.edu/globus.tar.gz',
                options=download_options)
        self.assertTrue(http_mock.called)
        req = http_mock.request_history[0]
        self.assertEqual(req.method, 'PUT', msg='not a PUT request first')
        self.assertEqual(os.path.basename(req.url), 'globus.tar.gz', msg='bad upload url')
        self.assertEqual(len(http_mock.request_history), 1, msg='more than one http request')

        # test bad upload
        http_mock.put('/globus.tar.gz', content=b'', headers={'ETAG': 'blah'})
        with self.assertRaises(Exception):
            await iceprod.core.functions.upload(filename,
                    'http://prod-exe.icecube.wisc.edu/globus.tar.gz',
                    options=download_options)

    @unittest_reporter(name='upload() file')
    async def test_404_upload(self):
        """Test the upload function"""
        data = 'the data'

        # upload file to local file system
        filename = os.path.join(self.test_dir, 'generators.py')
        out_dir = os.path.join(self.test_dir, 'output')
        os.makedirs(out_dir)
        output_file = os.path.join(out_dir, 'generators.py')
        with open(filename, 'w') as f:
            f.write(data)
        await iceprod.core.functions.upload('file:'+filename, 'file:'+output_file)
        if not os.path.isfile(output_file):
            raise Exception('copied file does not exist')
        data2 = open(output_file).read()
        self.assertEqual(data2, data, msg='data not equal')

        # test overwriting dest
        with open(output_file, 'w') as f:
            f.write('other data')
        await iceprod.core.functions.upload('file:'+filename, 'file:'+output_file)
        if not os.path.isfile(output_file):
            raise Exception('copied file does not exist')
        data2 = open(output_file).read()
        self.assertEqual(data2, data, msg='data not equal')

    @patch('iceprod.core.functions.GridFTP')
    @unittest_reporter(name='upload() gridftp')
    async def test_405_upload(self, gridftp):
        """Test the upload function"""
        data = 'the data'
        sha512sum = '8580e83fc859a2786430406fd41c7c6a0d3ac77b7eff07bc94c880f5b6e86b87320ea25cb3f3c5a3881236cf8bda92cb8f61c2a813881fee1d8f8331565ce98a'
        
        def put(url,filename=None):
            logger.info('fake get: url=%r, filename=%r', url, filename)
            if url.endswith('globus.tar.gz'):
                put.url = url
                put.filename = filename
            elif url.endswith('globus2.tar.gz'):
                raise Exception('expected failure')
            else:
                raise Exception()
        put.url = None
        put.filename = None
        gridftp.put = put
        gridftp.sha512sum.return_value = sha512sum

        filename = os.path.join(self.test_dir, 'globus.tar.gz')
        with open(filename, 'w') as f:
            f.write(data)
        await iceprod.core.functions.upload(filename,
                'gsiftp://data.icecube.wisc.edu/data/sim/sim-new/downloads/globus.tar.gz')
        self.assertEqual(put.url, 'gsiftp://data.icecube.wisc.edu/data/sim/sim-new/downloads/globus.tar.gz')
        self.assertEqual(put.filename, filename)

        # test gridftp error
        with self.assertRaises(Exception):
            await iceprod.core.functions.upload(filename,
                    'gsiftp://data.icecube.wisc.edu/data/sim/sim-new/downloads/globus2.tar.gz')

        # test checksum error
        gridftp.sha512sum.return_value = 'blah'
        with self.assertRaises(Exception):
            await iceprod.core.functions.upload(filename,
                    'gsiftp://data.icecube.wisc.edu/data/sim/sim-new/downloads/globus.tar.gz')

    @unittest_reporter(name='upload() dir')
    async def test_410_upload(self):
        """Test the upload function"""
        data = 'the data'

        # upload file to local file system
        in_dir = os.path.join(self.test_dir, 'input')
        filename = os.path.join(in_dir, 'generators.py')
        out_dir = os.path.join(self.test_dir, 'output')
        os.makedirs(in_dir)
        os.makedirs(out_dir)
        output_file = os.path.join(out_dir, 'generators')
        with open(filename, 'w') as f:
            f.write(data)
        await iceprod.core.functions.upload(in_dir, 'file:'+output_file)
        if not os.path.isfile(output_file):
            raise Exception('tar file does not exist')
        os.chdir(out_dir)
        subprocess.check_call(['tar','-axf',output_file])
        final_out = os.path.join('input','generators.py')
        if not os.path.isfile(final_out):
            raise Exception('copied file does not exist')
        data2 = open(final_out).read()
        self.assertEqual(data2, data, msg='data not equal')

    @unittest_reporter(name='upload() errors')
    async def test_420_upload(self):
        # bad request type
        filename = os.path.join(self.test_dir, 'globus.tar.gz')
        with self.assertRaises(Exception):
            await iceprod.core.functions.upload(filename,
                    'foobar://data.icecube.wisc.edu/data/sim/sim-new/downloads/globus2.tar.gz')

        # src doesn't exist
        filename = os.path.join(self.test_dir, 'globus.tar.gz')
        with self.assertRaises(Exception):
            await iceprod.core.functions.upload(filename,
                    'gsiftp://data.icecube.wisc.edu/data/sim/sim-new/downloads/globus2.tar.gz')

    @requests_mock.mock()
    @unittest_reporter(name='delete() http')
    def test_503_delete(self, http_mock):
        """Test the delete function"""
        download_options = {'username': 'user',
                            'password': 'pass',}
        data = 'the data'

        # delete file from http
        http_mock.delete('/globus.tar.gz', content=b'')
        iceprod.core.functions.delete('http://prod-exe.icecube.wisc.edu/globus.tar.gz',
                options=download_options)
        self.assertTrue(http_mock.called)
        req = http_mock.request_history[0]
        self.assertEqual(req.method, 'DELETE', msg='not a DELETE request')
        self.assertEqual(os.path.basename(req.url), 'globus.tar.gz', msg='bad delete url')

        # test http error
        try:
            iceprod.core.functions.delete('http://prod-exe.icecube.wisc.edu/globus2.tar.gz')
        except:
            pass
        else:
            raise Exception('did not raise Exception')

    @unittest_reporter(name='delete() file')
    def test_504_delete(self):
        """Test the delete function"""
        data = 'the data'

        # delete file from local file system
        filename = os.path.join(self.test_dir, 'generators.py')
        with open(filename, 'w') as f:
            f.write(data)
        iceprod.core.functions.delete(filename)
        if os.path.isfile(filename):
            raise Exception('delete file exists')

        # test with file: prefix
        with open(filename, 'w') as f:
            f.write(data)
        iceprod.core.functions.delete('file:'+filename)
        if os.path.isfile(filename):
            raise Exception('delete file exists')

        # test with non-existent file
        iceprod.core.functions.delete('file:'+filename)
        if os.path.isfile(filename):
            raise Exception('delete file exists')

    @patch('iceprod.core.functions.GridFTP')
    @unittest_reporter(name='delete() gridftp')
    def test_505_delete(self, gridftp):
        """Test the delete function"""
        data = 'the data'
        sha512sum = '8580e83fc859a2786430406fd41c7c6a0d3ac77b7eff07bc94c880f5b6e86b87320ea25cb3f3c5a3881236cf8bda92cb8f61c2a813881fee1d8f8331565ce98a'
        
        def delete(url):
            logger.info('fake delete: url=%r', url)
            if url.endswith('globus.tar.gz'):
                return
            elif url.endswith('globus2.tar.gz'):
                raise Exception('expected failure')
            else:
                raise Exception()
        gridftp.rmtree = delete

        url = 'gsiftp://data.icecube.wisc.edu/data/sim/sim-new/downloads/globus.tar.gz'
        iceprod.core.functions.delete(url)

        # test gridftp error
        try:
            url = 'gsiftp://data.icecube.wisc.edu/data/sim/sim-new/downloads/globus2.tar.gz'
            iceprod.core.functions.delete(url)
        except:
            pass
        else:
            raise Exception('did not raise Exception')

        # test gridftp error
        try:
            url = 'gsiftp://data.icecube.wisc.edu/data/sim/sim-new/downloads/blah.tar.gz'
            iceprod.core.functions.delete(url)
        except:
            pass
        else:
            raise Exception('did not raise Exception')

    @unittest_reporter(name='delete() errors')
    def test_510_delete(self):
        try:
            url = 'blah://test.test'
            iceprod.core.functions.delete(url)
        except:
            pass
        else:
            raise Exception('did not raise Exception')

def load_tests(loader, tests, pattern):
    suite = unittest.TestSuite()
    alltests = glob_tests(loader.getTestCaseNames(functions_test))
    suite.addTests(loader.loadTestsFromNames(alltests,functions_test))
    return suite



import pytest
import requests.exceptions

import iceprod.server.states


async def test_rest_datasets_empty(server):
    client = server(roles=['system'])

    ret = await client.request('GET', '/datasets')
    assert ret == {}


async def test_rest_datasets_post(server):
    client = server(roles=['user'])

    data = {
        'description': 'blah',
        'tasks_per_job': 4,
        'jobs_submitted': 1,
        'tasks_submitted': 4,
        'group': 'users',
    }
    ret = await client.request('POST', '/datasets', data)
    dataset_id = ret['result']

    ret = await client.request('GET', '/datasets')
    assert dataset_id in ret

    ret = await client.request('GET', '/datasets', {'status': 'suspended'})
    assert dataset_id not in ret

    ret = await client.request('GET', '/datasets', {'keys': 'dataset_id|dataset'})
    assert dataset_id in ret
    assert set(ret[dataset_id].keys()) == set(['dataset_id', 'dataset'])

    ret = await client.request('GET', f'/datasets/{dataset_id}')
    for k in data:
        assert k in ret
        assert data[k] == ret[k]


async def test_rest_datasets_post_system(server):
    client = server(roles=['system'])

    data = {
        'description': 'blah',
        'tasks_per_job': 4,
        'jobs_submitted': 1,
        'tasks_submitted': 4,
        'group': 'users',
        'username': 'fbar'
    }
    ret = await client.request('POST', '/datasets', data)
    dataset_id = ret['result']

    ret = await client.request('GET', f'/datasets/{dataset_id}')
    for k in data:
        assert k in ret
        assert data[k] == ret[k]


async def test_rest_datasets_post_always_active(server):
    client = server(roles=['user'])

    data = {
        'description': 'blah',
        'tasks_per_job': 4,
        'jobs_submitted': 1,
        'tasks_submitted': 4,
        'group': 'users',
        'always_active': True,
    }
    ret = await client.request('POST', '/datasets', data)
    dataset_id = ret['result']

    ret = await client.request('GET', '/datasets')
    assert dataset_id in ret

    ret = await client.request('GET', f'/datasets/{dataset_id}')
    for k in data:
        assert k in ret
        assert data[k] == ret[k]


async def test_rest_datasets_err(server):
    client = server(roles=['system'])
    with pytest.raises(requests.exceptions.HTTPError) as exc_info:
        await client.request('GET', '/datasets/bar')
    assert exc_info.value.response.status_code == 404


async def test_rest_datasets_update_description_user(server):
    client = server(roles=['user'], groups=['users'])

    data = {
        'description': 'blah',
        'tasks_per_job': 4,
        'jobs_submitted': 1,
        'tasks_submitted': 4,
        'group': 'users',
    }
    ret = await client.request('POST', '/datasets', data)
    dataset_id = ret['result']

    data = {'description': 'foo bar baz'}
    await client.request('PUT', f'/datasets/{dataset_id}/description', data)

    ret = await client.request('GET', f'/datasets/{dataset_id}')
    assert data['description'] == ret['description']


async def test_rest_datasets_update_description_group(server):
    client = server(roles=['user'], groups=['simprod','users'])

    data = {
        'description': 'blah',
        'tasks_per_job': 4,
        'jobs_submitted': 1,
        'tasks_submitted': 4,
        'group': 'simprod',
    }
    ret = await client.request('POST', '/datasets', data)
    dataset_id = ret['result']

    client = server(username='foo', roles=['user'], groups=['simprod','users'])
    data = {'description': 'foo bar baz'}
    await client.request('PUT', f'/datasets/{dataset_id}/description', data)

    ret = await client.request('GET', f'/datasets/{dataset_id}')
    assert data['description'] == ret['description']


async def test_rest_datasets_update_description_bad_user(server):
    client = server(roles=['user'], groups=['users'])

    data = {
        'description': 'blah',
        'tasks_per_job': 4,
        'jobs_submitted': 1,
        'tasks_submitted': 4,
        'group': 'users',
    }
    ret = await client.request('POST', '/datasets', data)
    dataset_id = ret['result']

    client = server(username='foo', roles=['user'], groups=['users'])
    data = {'description': 'foo bar baz'}
    with pytest.raises(requests.exceptions.HTTPError) as exc_info:
        await client.request('PUT', f'/datasets/{dataset_id}/description', data)
    assert exc_info.value.response.status_code == 403


async def test_rest_datasets_update_status(server):
    client = server(roles=['user'], groups=['users'])

    data = {
        'description': 'blah',
        'tasks_per_job': 4,
        'jobs_submitted': 1,
        'tasks_submitted': 4,
        'group': 'users',
    }
    ret = await client.request('POST', '/datasets', data)
    dataset_id = ret['result']

    data = {'status': 'suspended'}
    await client.request('PUT', f'/datasets/{dataset_id}/status', data)

    ret = await client.request('GET', f'/datasets/{dataset_id}')
    assert data['status'] == ret['status']


async def test_rest_datasets_update_priority(server):
    client = server(roles=['user'], groups=['users'])

    data = {
        'description': 'blah',
        'tasks_per_job': 4,
        'jobs_submitted': 1,
        'tasks_submitted': 4,
        'priority': .5,
        'group': 'users',
    }
    ret = await client.request('POST', '/datasets', data)
    dataset_id = ret['result']

    data = {'priority': .8}
    await client.request('PUT', f'/datasets/{dataset_id}/priority', data)

    ret = await client.request('GET', f'/datasets/{dataset_id}')
    assert data['priority'] == ret['priority']


async def test_rest_datasets_hard_reset(server):
    client = server(roles=['user'], groups=['users'])

    data = {
        'description': 'blah',
        'tasks_per_job': 4,
        'jobs_submitted': 1,
        'tasks_submitted': 4,
        'group': 'users',
        'status': 'complete',
    }
    ret = await client.request('POST', '/datasets', data)
    dataset_id = ret['result']

    await client.request('POST', f'/datasets/{dataset_id}/dataset_actions/hard_reset', {})

    ret = await client.request('GET', f'/datasets/{dataset_id}')
    assert ret['status'] == iceprod.server.states.DATASET_STATUS_START


async def test_rest_datasets_truncate(server):
    client = server(roles=['user'], groups=['users'])

    data = {
        'description': 'blah',
        'tasks_per_job': 4,
        'jobs_submitted': 1,
        'tasks_submitted': 4,
        'group': 'users',
    }
    ret = await client.request('POST', '/datasets', data)
    dataset_id = ret['result']

    await client.request('POST', f'/datasets/{dataset_id}/dataset_actions/truncate', {})

    ret = await client.request('GET', f'/datasets/{dataset_id}')
    assert ret['truncated'] is True


async def test_rest_datasets_update_jobs_submitted(server):
    client = server(roles=['user'], groups=['users'])

    data = {
        'description': 'blah',
        'tasks_per_job': 4,
        'jobs_submitted': 1,
        'tasks_submitted': 4,
        'group': 'users',
    }
    ret = await client.request('POST', '/datasets', data)
    dataset_id = ret['result']

    data = {'jobs_submitted': 2}
    await client.request('PUT', f'/datasets/{dataset_id}/jobs_submitted', data)

    ret = await client.request('GET', f'/datasets/{dataset_id}')
    assert data['jobs_submitted'] == ret['jobs_submitted']

    with pytest.raises(requests.exceptions.HTTPError) as exc_info:
        await client.request('PUT', f'/datasets/{dataset_id}/jobs_submitted')
    assert exc_info.value.response.status_code == 400

    data = {}
    with pytest.raises(requests.exceptions.HTTPError) as exc_info:
        await client.request('PUT', f'/datasets/{dataset_id}/jobs_submitted', data)
    assert exc_info.value.response.status_code == 400

    
    data = {'jobs_submitted': 'foo'}
    with pytest.raises(requests.exceptions.HTTPError) as exc_info:
        await client.request('PUT', f'/datasets/{dataset_id}/jobs_submitted', data)
    assert exc_info.value.response.status_code == 400

    data = {'jobs_submitted': 0}
    with pytest.raises(requests.exceptions.HTTPError) as exc_info:
        await client.request('PUT', f'/datasets/{dataset_id}/jobs_submitted', data)
    assert exc_info.value.response.status_code == 400

    data = {
        'description': 'blah',
        'tasks_per_job': 4,
        'jobs_submitted': 1,
        'tasks_submitted': 4,
        'group': 'users',
        'jobs_immutable': True,
    }
    ret = await client.request('POST', '/datasets', data)
    dataset_id = ret['result']

    data = {'jobs_submitted': 2}
    with pytest.raises(requests.exceptions.HTTPError) as exc_info:
        await client.request('PUT', f'/datasets/{dataset_id}/jobs_submitted', data)
    assert exc_info.value.response.status_code == 400


async def test_rest_datasets_summaries_status(server):
    client = server(roles=['user'], groups=['users'])

    data = {
        'description': 'blah',
        'tasks_per_job': 4,
        'jobs_submitted': 1,
        'tasks_submitted': 4,
        'group': 'users',
    }
    ret = await client.request('POST', '/datasets', data)
    dataset_id = ret['result']

    ret = await client.request('GET', '/dataset_summaries/status')
    assert ret == {'processing': [dataset_id]}

    client2 = server(roles=['user'], groups=['filtering'], username='user2')
    ret = await client2.request('GET', '/dataset_summaries/status')
    assert ret == {'processing': [dataset_id]}


async def test_rest_datasets_summaries_status_with_groups(server):
    client = server(roles=['user'], groups=['simprod'])

    data = {
        'description': 'blah',
        'tasks_per_job': 4,
        'jobs_submitted': 1,
        'tasks_submitted': 4,
        'group': 'simprod',
        'auth_groups_read': [],
    }
    ret = await client.request('POST', '/datasets', data)
    dataset_id = ret['result']

    ret = await client.request('GET', '/dataset_summaries/status')
    assert ret == {'processing': [dataset_id]}

    client2 = server(roles=['user'], groups=['filtering'], username='user2')
    ret = await client2.request('GET', '/dataset_summaries/status')
    assert ret == {}



#!/usr/bin/env python3

"""
This is a custom HTCondor file transfer plugin for GridFTP.
In this example, it transfers files described by a gridftp://path/to/file URL
by copying them from the path indicated to a job's working directory.
"""

import glob
import os
import sys
import subprocess
import time

DEFAULT_TIMEOUT = 300
PLUGIN_VERSION = '1.0.0'

EXIT_SUCCESS = 0
EXIT_FAILURE = 1
EXIT_AUTHENTICATION_REFRESH = 2

try:
    from classad import ClassAd, parseAds  # type: ignore
except ImportError:
    import re
    import json

    class ClassAd(dict):  # type: ignore
        def printOld(self):
            ret = []
            for k,v in self.items():
                if isinstance(v, str):
                    v = '"{0}"'.format(v)
                ret.append('{0} = {1}'.format(k, v))
            return '[\n' + ';\n'.join(ret) + '\n]\n'

    def parseAds(data):
        ret = []
        for ad in re.findall(r'\s*?\[([\w\W]*?)\]', data):
            ad_ret = {}
            for k,v in re.findall(r'\s*?(\w+) += +([^;\n]+);?', ad):
                ad_ret[k] = json.loads(v)
            ret.append(ad_ret)
        return ret


def print_help(stream=sys.stderr):
    help_msg = '''Usage: {0} -infile <input-filename> -outfile <output-filename>
       {0} -classad

Options:
  -classad                    Print a ClassAd containing the capablities of this
                              file transfer plugin.
  -infile <input-filename>    Input ClassAd file
  -outfile <output-filename>  Output ClassAd file
  -upload                     Indicates this transfer is an upload (default is
                              download)
'''
    stream.write(help_msg.format(sys.argv[0]))


def print_capabilities():
    capabilities = {
        'MultipleFileSupport': True,
        'PluginType': 'FileTransfer',
        # SupportedMethods indicates which URL methods/types this plugin supports
        'SupportedMethods': 'gsiftp',
        'Version': PLUGIN_VERSION,
    }
    sys.stdout.write(ClassAd(capabilities).printOld())


def parse_args():

    # The only argument lists that are acceptable are
    # <this> -classad
    # <this> -infile <input-filename> -outfile <output-filename>
    # <this> -outfile <output-filename> -infile <input-filename>
    if len(sys.argv) not in [2, 5, 6]:
        print_help()
        sys.exit(EXIT_FAILURE)

    # If -classad, print the capabilities of the plugin and exit early
    if (len(sys.argv) == 2) and (sys.argv[1] == '-classad'):
        print_capabilities()
        sys.exit(EXIT_SUCCESS)

    # If -upload, set is_upload to True and remove it from the args list
    is_upload = False
    if '-upload' in sys.argv[1:]:
        is_upload = True
        sys.argv.remove('-upload')

    # -infile and -outfile must be in the first and third position
    if not (
            ('-infile' in sys.argv[1:]) and
            ('-outfile' in sys.argv[1:]) and
            (sys.argv[1] in ['-infile', '-outfile']) and
            (sys.argv[3] in ['-infile', '-outfile']) and
            (len(sys.argv) == 5)):
        print_help()
        sys.exit(1)
    infile = None
    outfile = None
    try:
        for i, arg in enumerate(sys.argv):
            if i == 0:
                continue
            elif arg == '-infile':
                infile = sys.argv[i+1]
            elif arg == '-outfile':
                outfile = sys.argv[i+1]
    except IndexError:
        print_help()
        sys.exit(EXIT_FAILURE)

    return {'infile': infile, 'outfile': outfile, 'upload': is_upload}


def format_error(error):
    return '{0}: {1}'.format(type(error).__name__, str(error))


def get_error_dict(error, url=''):
    error_string = format_error(error)
    error_dict = {
        'TransferSuccess': False,
        'TransferError': error_string,
        'TransferUrl': url,
    }

    return error_dict


class GridftpPlugin:

    def setup_env(self):
        # if a proxy exists, use it
        proxies = glob.glob(os.path.join(os.getcwd(), 'x509up_*'))
        if proxies:
            os.environ['X509_USER_PROXY'] = proxies[0]
        else:
            raise RuntimeError('X509_USER_PROXY does not exist')

        if not os.path.exists('/cvmfs/icecube.opensciencegrid.org/iceprod/v2.7.1/env-shell.sh'):
            raise RuntimeError('CVMFS does not exist')

    def _do_transfer(self, inpath, outpath):
        try:
            subprocess.check_output([
                '/cvmfs/icecube.opensciencegrid.org/iceprod/v2.7.1/env-shell.sh',
                'globus-url-copy',
                '-cd',
                '-rst',
                inpath,
                outpath,
            ], stderr=subprocess.STDOUT)
        except subprocess.CalledProcessError as e:
            if e.output:
                output = e.output.decode('utf-8')
                for line in output.split('\n'):
                    if line.lower().startswith('error'):
                        raise RuntimeError('globus-url-copy failed: '+line)
                raise RuntimeError('Generic subprocess failure: '+output)
            raise

    def download_file(self, url, local_file_path):

        start_time = time.time()

        # Download transfer logic goes here
        self._do_transfer(url, 'file://'+os.path.abspath(local_file_path))
        file_size = os.stat(local_file_path).st_size

        end_time = time.time()

        # Get transfer statistics
        transfer_stats = {
            'TransferSuccess': True,
            'TransferProtocol': 'gsiftp',
            'TransferType': 'upload',
            'TransferFileName': local_file_path,
            'TransferFileBytes': file_size,
            'TransferTotalBytes': file_size,
            'TransferStartTime': int(start_time),
            'TransferEndTime': int(end_time),
            'ConnectionTimeSeconds': end_time - start_time,
            'TransferUrl': url,
        }

        return transfer_stats

    def upload_file(self, url, local_file_path):

        start_time = time.time()

        # Upload transfer logic goes here
        self._do_transfer('file://'+os.path.abspath(local_file_path), url)
        file_size = os.stat(local_file_path).st_size

        end_time = time.time()

        # Get transfer statistics
        transfer_stats = {
            'TransferSuccess': True,
            'TransferProtocol': 'gsiftp',
            'TransferType': 'upload',
            'TransferFileName': local_file_path,
            'TransferFileBytes': file_size,
            'TransferTotalBytes': file_size,
            'TransferStartTime': int(start_time),
            'TransferEndTime': int(end_time),
            'ConnectionTimeSeconds': end_time - start_time,
            'TransferUrl': url,
        }

        return transfer_stats


if __name__ == '__main__':

    # Start by parsing input arguments
    try:
        args = parse_args()
    except Exception:
        sys.exit(EXIT_FAILURE)

    gridftp_plugin = GridftpPlugin()

    # Parse in the classads stored in the input file.
    # Each ad represents a single file to be transferred.
    try:
        infile_ads = parseAds(open(args['infile'], 'r'))
    except Exception as err:
        try:
            with open(args['outfile'], 'w') as outfile:
                outfile_dict = get_error_dict(err)
                outfile.write(str(ClassAd(outfile_dict)))
        except Exception:
            pass
        sys.exit(EXIT_FAILURE)

    # Now iterate over the list of classads and perform the transfers.
    try:
        with open(args['outfile'], 'w') as outfile:
            for ad in infile_ads:
                try:
                    gridftp_plugin.setup_env()
                    if not args['upload']:
                        outfile_dict = gridftp_plugin.download_file(ad['Url'], ad['LocalFileName'])
                    else:
                        outfile_dict = gridftp_plugin.upload_file(ad['Url'], ad['LocalFileName'])

                    outfile.write(str(ClassAd(outfile_dict)))

                except Exception as err:
                    try:
                        outfile_dict = get_error_dict(err, url=ad['Url'])
                        outfile.write(str(ClassAd(outfile_dict)))
                    except Exception:
                        pass
                    sys.exit(EXIT_FAILURE)

    except Exception:
        sys.exit(EXIT_FAILURE)



"""
Test script for priority
"""
import datetime
import time
import logging
from unittest.mock import MagicMock

import pytest
import iceprod.server
from iceprod.server import priority


logger = logging.getLogger('priority_test')


@pytest.fixture(autouse=True)
def set_time(monkeypatch):
    now = datetime.datetime(2024, 1, 1, 1, 10, 0, 0, datetime.UTC)
    mock = MagicMock()
    mock.now = MagicMock(return_value=now)
    monkeypatch.setattr(iceprod.server.priority, 'datetime', mock)
    tnow = time.mktime(now.utctimetuple())
    tmock = MagicMock(return_value=tnow)
    monkeypatch.setattr(time, 'time', tmock)
    yield now


def prio_setup():
    datasets = {
        'd0': {
            'dataset_id': 'd0',
            'jobs_submitted': 10,
            'tasks_submitted': 20,
            'priority': 1,
            'start_date': '2024-01-01T01:00:00',
            'group': 'users',
            'username': 'u_a',
            'tasks': {
                't0': {'task_id': 't0', 'task_index': 0, 'job_index': 0},
                't1': {'task_id': 't1', 'task_index': 1, 'job_index': 0},
            }
        },
        'd1': {
            'dataset_id': 'd1',
            'jobs_submitted': 10,
            'tasks_submitted': 20,
            'priority': 1,
            'start_date': '2024-01-01T01:00:00',
            'group': 'users',
            'username': 'u_b',
            'tasks': {
                't2': {'task_id': 't2', 'task_index': 0, 'job_index': 0},
                't3': {'task_id': 't3', 'task_index': 1, 'job_index': 0},
            }
        },
        'd2': {
            'dataset_id': 'd2',
            'jobs_submitted': 1000,
            'tasks_submitted': 20000,
            'priority': 1,
            'start_date': '2024-01-01T01:00:00',
            'group': 'simprod',
            'username': 'u_a',
            'tasks': {
                't4': {'task_id': 't4', 'task_index': 0, 'job_index': 400},
                't5': {'task_id': 't5', 'task_index': 1, 'job_index': 400},
            }
        },
        'd3': {
            'dataset_id': 'd3',
            'jobs_submitted': 1000,
            'tasks_submitted': 20000,
            'priority': 1,
            'start_date': '2023-10-01T01:00:00',
            'group': 'simprod',
            'username': 'u_a',
            'tasks': {
                't6': {'task_id': 't6', 'task_index': 0, 'job_index': 400},
                't7': {'task_id': 't7', 'task_index': 1, 'job_index': 400},
            }
        },
    }
    users = {
        'u_a': {'username': 'u_a', 'priority': 1.},
        'u_b': {'username': 'u_b', 'priority': .5},
    }

    p = priority.Priority(None)
    p.dataset_cache = datasets
    p.user_cache = users
    return p


async def test_10_get_dataset_prio():
    """Test get_dataset_prio"""
    p = prio_setup()
    prio1 = await p.get_dataset_prio('d0')
    prio2 = await p.get_dataset_prio('d1')
    assert prio2 < prio1

    prio3 = await p.get_dataset_prio('d2')
    assert prio3 < prio1

    prio4 = await p.get_dataset_prio('d3')
    assert prio4 > prio3


async def test_20_get_task_prio():
    """Test get_task_prio"""
    p = prio_setup()
    prio1 = await p.get_task_prio('d0', 't0')
    prio2 = await p.get_task_prio('d1', 't2')
    # equal at 1.0 because of boost
    assert prio2 == prio1

    prio3 = await p.get_task_prio('d2', 't4')
    prio4 = await p.get_task_prio('d2', 't5')
    assert prio3 < prio1 
    assert prio3 < prio4  # ordering of tasks in a job



from collections import Counter
from dataclasses import dataclass
import logging
from pprint import pprint
from unittest.mock import MagicMock, AsyncMock

from prometheus_client import REGISTRY
import pytest
import requests.exceptions

from iceprod.util import VERSION_STRING
from iceprod.core.resources import Resources
from iceprod.core.config import Dataset, Job, Task
import iceprod.server.config
import iceprod.server.grid


def test_grid_init():
    override = ['queue.type=test', 'queue.site=bar']
    cfg = iceprod.server.config.IceProdConfig(save=False, override=override)

    rc = MagicMock()
    g = iceprod.server.grid.BaseGrid(cfg=cfg, rest_client=rc, cred_client=None)

    assert 'gpu' not in g.site_requirements

    pprint(list(REGISTRY.collect()))
    metric = REGISTRY.get_sample_value('iceprod_info', {
        'name': 'bar',
        'type': 'grid',
        'queue_type': 'test',
        'version': VERSION_STRING,
        'exclusive': 'False',
    })
    assert metric == 1

def test_grid_init_gpu():
    override = ['queue.type=test', 'queue.site=grid-gpu']
    cfg = iceprod.server.config.IceProdConfig(save=False, override=override)

    rc = MagicMock()
    g = iceprod.server.grid.BaseGrid(cfg=cfg, rest_client=rc, cred_client=None)

    assert 'gpu' in g.site_requirements

    pprint(list(REGISTRY.collect()))
    metric = REGISTRY.get_sample_value('iceprod_info', {
        'name': 'grid-gpu',
        'type': 'grid',
        'queue_type': 'test',
        'version': VERSION_STRING,
        'exclusive': 'False',
    })
    assert metric == 1

def test_grid_init_exclusive():
    override = ['queue.type=test', 'queue.site=grid', 'queue.exclusive=true']
    cfg = iceprod.server.config.IceProdConfig(save=False, override=override)

    rc = MagicMock()
    g = iceprod.server.grid.BaseGrid(cfg=cfg, rest_client=rc, cred_client=None)

    assert 'requirements.site' in g.site_query_params
    assert g.site_query_params['requirements.site'] == 'grid'

async def test_grid_run():
    override = ['queue.type=test']
    cfg = iceprod.server.config.IceProdConfig(save=False, override=override)

    rc = MagicMock()
    g = iceprod.server.grid.BaseGrid(cfg=cfg, rest_client=rc, cred_client=None)

    with pytest.raises(NotImplementedError):
        await g.run()


async def test_grid_dataset_lookup(monkeypatch):
    override = ['queue.type=test']
    cfg = iceprod.server.config.IceProdConfig(save=False, override=override)

    rc = MagicMock()
    g = iceprod.server.grid.BaseGrid(cfg=cfg, rest_client=rc, cred_client=None)

    dataset_mock = AsyncMock()
    d = dataset_mock.load_from_api.return_value
    d.fill_defaults = MagicMock()
    d.validate = MagicMock()
    monkeypatch.setattr(iceprod.server.grid, 'Dataset', dataset_mock)
    ret = await g.dataset_lookup('12345')
    assert ret == d
    assert dataset_mock.load_from_api.call_count == 1

    # test cache miss
    ret = await g.dataset_lookup('6789')
    assert ret == d
    assert dataset_mock.load_from_api.call_count == 2

    # test cache hit
    ret = await g.dataset_lookup('12345')
    assert ret == d
    assert dataset_mock.load_from_api.call_count == 2


@pytest.mark.parametrize('num_tasks,queue,avail,priorities,out_tasks', [
    (10, {}, {'d1': 50}, {'d1': 0.5}, {'d1': 10}),
    (10, {'d1': 100}, {'d1': 5}, {'d1': 0.5}, {'d1': 5}),
    (10, {}, {'d1': 2, 'd2': 3}, {'d1': 0.5, 'd2': 0.5}, {'d1': 2, 'd2': 3}),
    (10, {'d1': 100, 'd2': 80}, {'d1': 10, 'd2': 10, 'd3': 2}, {'d1': 0.5, 'd2': 0.5, 'd3': 0.5}, {'d2': 8, 'd3': 2}),
    (10, {'d1': 100, 'd2': 80}, {'d1': 10, 'd2': 10, 'd3': 20}, {'d1': 0.5, 'd2': 0.5, 'd3': 0.5}, {'d3': 10}),
    (10, {'d1': 10, 'd2': 21}, {'d1': 20, 'd2': 20}, {'d1': 0.25, 'd2': 0.5}, {'d1': 10}),
    (10, {'d1': 10, 'd2': 19}, {'d1': 20, 'd2': 20}, {'d1': 0.25, 'd2': 0.5}, {'d2': 10}),
])
async def test_grid_get_tasks_to_queue(num_tasks, queue, avail, priorities, out_tasks):
    override = ['queue.type=test', 'queue.check_time=0']
    cfg = iceprod.server.config.IceProdConfig(save=False, override=override)

    rc = MagicMock()
    g = iceprod.server.grid.BaseGrid(cfg=cfg, rest_client=rc, cred_client=None)

    async def rest_queue(method, path, args):
        num = args.get('num', 1)
        dataset_exclude = args.get('dataset_deprio', [])
        ret = []
        while len(ret) < num:
            for d in avail:
                if avail[d] > 0 and d not in dataset_exclude:
                    avail[d] -= 1
                    ret.append({'dataset_id': d})
                    break
            else:
                if dataset_exclude:
                    dataset_exclude.pop()
                else:
                    break
        if ret:
            return ret
        r = MagicMock()
        r.status_code = 404
        logging.info('404!')
        raise requests.exceptions.HTTPError(response=r)
    
    def convert(val):
        ret = MagicMock()
        ret.dataset.dataset_id = val['dataset_id']
        return ret

    rc.request = AsyncMock(side_effect=rest_queue)
    g._convert_to_task = AsyncMock(side_effect=convert)
    g._get_resources = MagicMock()
    g._get_priority_object = AsyncMock()
    g._get_priority_object.return_value.get_dataset_prio = AsyncMock(side_effect=lambda d: priorities[d])
    g.queue_dataset_status = MagicMock(return_value={iceprod.server.grid.GridStatus.QUEUED: queue})

    tasks = await g.get_tasks_to_queue(num_tasks)

    out_count = sum(out_tasks.values())

    assert len(tasks) == out_count
    assert Counter(t.dataset.dataset_id for t in tasks) == out_tasks
    assert g._convert_to_task.call_count == out_count
    assert g._get_resources.call_count == out_count


async def test_grid_submit_none():
    override = ['queue.type=test', 'queue.check_time=0']
    cfg = iceprod.server.config.IceProdConfig(save=False, override=override)

    rc = MagicMock()
    g = iceprod.server.grid.BaseGrid(cfg=cfg, rest_client=rc, cred_client=None)

    NUM_TASKS = 2
    response = MagicMock()
    response.status_code = 404
    rc.request = AsyncMock(side_effect=requests.exceptions.HTTPError(response=response))
    g._convert_to_task = AsyncMock()
    g._get_resources = MagicMock()
    g._get_priority_object = AsyncMock()
    g._get_priority_object.return_value.get_dataset_prio = AsyncMock(return_value=0.5)
    g.queue_dataset_status = MagicMock(return_value={})

    tasks = await g.get_tasks_to_queue(NUM_TASKS)

    assert len(tasks) == 0
    assert rc.request.call_count == 1
    assert g._convert_to_task.call_count == 0
    assert g._get_resources.call_count == 0


async def test_grid_convert_to_task(monkeypatch):
    override = ['queue.type=test', 'queue.check_time=0', 'queue.site_temp=http://foo.bar']
    cfg = iceprod.server.config.IceProdConfig(save=False, override=override)

    rc = MagicMock()
    rc.request = AsyncMock()
    g = iceprod.server.grid.BaseGrid(cfg=cfg, rest_client=rc, cred_client=None)

    TASK = MagicMock()
    DATASET = MagicMock()
    DATASET.config = {'options':{}, 'version': 3.2}
    g.dataset_lookup = AsyncMock(return_value=DATASET)
    rc.request.return_value = {'files': []}

    ret = await g._convert_to_task(TASK)

    assert ret.dataset.config['options']['site_temp'] == 'http://foo.bar'


async def test_grid_get_dataset_credentials():
    override = ['queue.type=test', 'queue.check_time=0', 'queue.site_temp=http://foo.bar', 'oauth_condor_client_id=client']
    cfg = iceprod.server.config.IceProdConfig(save=False, override=override)

    rc = MagicMock()
    rc.request = AsyncMock()
    cred_client = MagicMock()
    cred_client.request = AsyncMock(return_value=['creds'])
    g = iceprod.server.grid.BaseGrid(cfg=cfg, rest_client=rc, cred_client=cred_client)

    d = Dataset(
        dataset_id='1234',
        dataset_num=1234,
        jobs_submitted=10,
        tasks_submitted=10,
        tasks_per_job=1,
        status='processing',
        priority=.5,
        group='group',
        user='user',
        debug=False,
        config={'tasks':[{'task_scopes':{}}]},
    )
    j = Job(dataset=d, job_id='5678', job_index=3, status='processing')
    t = Task(
        dataset=d,
        job=j,
        task_id='91011',
        task_index=0,
        name='generate',
        depends=[],
        requirements={},
        status='queued',
        site='site',
        stats={},
    )

    ret = await g._get_dataset_credentials(t)
    assert ret == ['creds']
    assert cred_client.request.call_count == 1
    assert cred_client.request.call_args == [('GET', '/datasets/1234/tasks/generate/exchange', {'client_id': 'client'}),]

    cred_client.request = AsyncMock(return_value=['creds2'])
    ret = await g._get_dataset_credentials(t)
    assert ret == ['creds']
    assert cred_client.request.call_count == 0

    
def test_grid_get_resources(i3prod_path):
    d = Dataset(
        dataset_id='1234',
        dataset_num=1234,
        jobs_submitted=10,
        tasks_submitted=10,
        tasks_per_job=1,
        status='processing',
        priority=.5,
        group='group',
        user='user',
        debug=False,
        config={},
    )
    j = Job(dataset=d, job_id='5678', job_index=3, status='processing')
    t = Task(
        dataset=d,
        job=j,
        task_id='91011',
        task_index=0,
        name='generate',
        depends=[],
        requirements={},
        status='queued',
        site='site',
        stats={},
    )

    r = iceprod.server.grid.BaseGrid._get_resources(t)
    assert list(r.keys()) == list(Resources.defaults.keys())

    t.requirements['cpu'] = 2
    r = iceprod.server.grid.BaseGrid._get_resources(t)
    assert r['cpu'] == 2

    t.requirements['gpu'] = 2
    r = iceprod.server.grid.BaseGrid._get_resources(t)
    assert r['gpu'] == 2

    t.requirements['os'] = 'RHEL_7_x86_64'
    r = iceprod.server.grid.BaseGrid._get_resources(t)
    assert r['os'] == ['RHEL_7_x86_64']  # type: ignore

    t.requirements['os'] = ['RHEL_7_x86_64', 'RHEL_8_x86_64']
    r = iceprod.server.grid.BaseGrid._get_resources(t)
    assert r['os'] == ['RHEL_7_x86_64', 'RHEL_8_x86_64']  # type: ignore


@dataclass(kw_only=True, slots=True)
class GT(iceprod.server.grid.GridTask):
    dataset_id: str  # type: ignore
    task_id: str  # type: ignore
    instance_id: str | None = None


async def test_grid_upload_log():
    override = ['queue.type=test']
    cfg = iceprod.server.config.IceProdConfig(save=False, override=override)

    rc = MagicMock()
    rc.request = AsyncMock()
    g = iceprod.server.grid.BaseGrid(cfg=cfg, rest_client=rc, cred_client=None)

    task = GT(dataset_id='ddd', task_id='ttt')
    name = 'name'
    data = 'the log\ndata'
    await g._upload_log(task, name=name, data=data)

    assert rc.request.call_count == 1

    rc.request = AsyncMock(side_effect=requests.exceptions.HTTPError())
    await g._upload_log(task, name=name, data=data)


async def test_grid_upload_stats():
    override = ['queue.type=test']
    cfg = iceprod.server.config.IceProdConfig(save=False, override=override)

    rc = MagicMock()
    rc.request = AsyncMock()
    g = iceprod.server.grid.BaseGrid(cfg=cfg, rest_client=rc, cred_client=None)

    task = GT(dataset_id='ddd', task_id='ttt')
    stats = {}
    await g._upload_stats(task, stats=stats)

    assert rc.request.call_count == 1
    
    stats = {
        'site': 'Foo'
    }
    await g._upload_stats(task, stats=stats)

    assert rc.request.call_args.args[-1]['site'] == 'Foo'

    rc.request = AsyncMock(side_effect=requests.exceptions.HTTPError())
    await g._upload_stats(task, stats=stats)


async def test_grid_task_idle():
    override = ['queue.type=test']
    cfg = iceprod.server.config.IceProdConfig(save=False, override=override)

    rc = MagicMock()
    rc.request = AsyncMock()
    g = iceprod.server.grid.BaseGrid(cfg=cfg, rest_client=rc, cred_client=None)

    task = GT(dataset_id='ddd', task_id='ttt', instance_id='iii')
    await g.task_idle(task)

    assert rc.request.call_count == 1

    response = MagicMock()
    response.status_code = 404
    rc.request = AsyncMock(side_effect=requests.exceptions.HTTPError(response=response))
    await g.task_idle(task)

    response.status_code = 500
    with pytest.raises(requests.exceptions.HTTPError):
        await g.task_idle(task)


async def test_grid_task_processing():
    override = ['queue.type=test']
    cfg = iceprod.server.config.IceProdConfig(save=False, override=override)

    rc = MagicMock()
    rc.request = AsyncMock()
    g = iceprod.server.grid.BaseGrid(cfg=cfg, rest_client=rc, cred_client=None)

    task = GT(dataset_id='ddd', task_id='ttt', instance_id='iii')
    await g.task_processing(task)

    assert rc.request.call_count == 1

    response = MagicMock()
    response.status_code = 404
    rc.request = AsyncMock(side_effect=requests.exceptions.HTTPError(response=response))
    await g.task_processing(task)

    response.status_code = 500
    with pytest.raises(requests.exceptions.HTTPError):
        await g.task_processing(task)


async def test_grid_task_reset():
    override = ['queue.type=test']
    cfg = iceprod.server.config.IceProdConfig(save=False, override=override)

    rc = MagicMock()
    rc.request = AsyncMock()
    g = iceprod.server.grid.BaseGrid(cfg=cfg, rest_client=rc, cred_client=None)

    task = GT(dataset_id='ddd', task_id='ttt', instance_id='iii')
    await g.task_reset(task)

    assert rc.request.call_count == 1

    response = MagicMock()
    response.status_code = 404
    rc.request = AsyncMock(side_effect=requests.exceptions.HTTPError(response=response))
    await g.task_reset(task, reason='reason')

    assert rc.request.call_args.args[-1]['reason'] == 'reason'

    response.status_code = 500
    with pytest.raises(requests.exceptions.HTTPError):
        await g.task_reset(task)


async def test_grid_task_failure(i3prod_path):
    override = ['queue.type=test']
    cfg = iceprod.server.config.IceProdConfig(save=False, override=override)

    rc = MagicMock()
    rc.request = AsyncMock()
    g = iceprod.server.grid.BaseGrid(cfg=cfg, rest_client=rc, cred_client=None)

    task = GT(dataset_id='ddd', task_id='ttt', instance_id='iii')
    await g.task_failure(task)

    assert rc.request.call_count == 1

    response = MagicMock()
    response.status_code = 404
    rc.request = AsyncMock(side_effect=requests.exceptions.HTTPError(response=response))
    await g.task_failure(task)

    response.status_code = 500
    with pytest.raises(requests.exceptions.HTTPError):
        await g.task_failure(task)

    rc.request = AsyncMock()
    await g.task_failure(task, reason='reason')
    assert rc.request.call_count == 2
    assert rc.request.call_args_list[0].args[-1]['reason'] == 'reason'

    rc.request = AsyncMock()
    stats = {'resources': {'cpu': 1}}
    await g.task_failure(task, stats=stats)
    assert rc.request.call_count == 2

    rc.request = AsyncMock()
    outfile = i3prod_path / 'outfile'
    outfile.write_text('out data')
    await g.task_failure(task, stdout=outfile)
    assert rc.request.call_count == 2

    rc.request = AsyncMock()
    outfile = i3prod_path / 'outfile'
    outfile.write_text('out data')
    errfile = i3prod_path / 'errfile'
    errfile.write_text('err message')
    await g.task_failure(task, stdout=outfile, stderr=errfile, reason='reason')
    assert rc.request.call_count == 4


async def test_grid_task_success(i3prod_path):
    override = ['queue.type=test']
    cfg = iceprod.server.config.IceProdConfig(save=False, override=override)

    rc = MagicMock()
    rc.request = AsyncMock()
    g = iceprod.server.grid.BaseGrid(cfg=cfg, rest_client=rc, cred_client=None)

    task = GT(dataset_id='ddd', task_id='ttt', instance_id='iii')
    await g.task_success(task)

    assert rc.request.call_count == 1

    response = MagicMock()
    response.status_code = 404
    rc.request = AsyncMock(side_effect=requests.exceptions.HTTPError(response=response))
    await g.task_success(task)

    response.status_code = 500
    with pytest.raises(requests.exceptions.HTTPError):
        await g.task_success(task)

    rc.request = AsyncMock()
    stats = {'resources': {'cpu': 1, 'time': 4.3}, 'site': 'MySite'}
    await g.task_success(task, stats=stats)
    assert rc.request.call_count == 2

    rc.request = AsyncMock()
    outfile = i3prod_path / 'outfile'
    outfile.write_text('out data')
    await g.task_success(task, stdout=outfile)
    assert rc.request.call_count == 2

    rc.request = AsyncMock()
    outfile = i3prod_path / 'outfile'
    outfile.write_text('out data')
    errfile = i3prod_path / 'errfile'
    errfile.write_text('err message')
    await g.task_success(task, stdout=outfile, stderr=errfile)
    assert rc.request.call_count == 3





#!/usr/bin/env python

import os
import subprocess
import argparse
import json
import glob
import logging
import asyncio
from collections import defaultdict

from rest_tools.client import RestClient

async def get_stats(args):
    rpc = RestClient('https://iceprod2-api.icecube.wisc.edu', args['token'])

    datasets = await rpc.request('GET', '/datasets', {'keys': 'dataset_id|dataset'})
    dataset_id = None
    for d in datasets:
        if datasets[d]['dataset'] == args['dataset']:
            dataset_id = d
            break
    else:
        raise Exception('bad dataset num')
    dataset = await rpc.request('GET', f'/datasets/{dataset_id}')
    jobs = await rpc.request('GET', f'/datasets/{dataset_id}/jobs', {'keys': 'job_id|job_index'})

    params = {
        'status': 'complete',
        'keys': 'task_id|name|job_id', 
    }
    tasks = await rpc.request('GET', f'/datasets/{dataset_id}/tasks', params)
    if not tasks:
        raise Exception('no tasks found')

    dataset_stats = defaultdict(dict)
    try:
        futures = []
        for task_id in sorted(tasks, key=lambda tid: jobs[tasks[tid]['job_id']]['job_index']):
            futures.append(asyncio.create_task(rpc.request('GET', f'/datasets/{dataset_id}/tasks/{task_id}/task_stats')))

        print(len(futures))
        for f in asyncio.as_completed(futures):
            raw_stats = await f
            stats = [raw_stats[k] for k in raw_stats if 'error_summary' not in raw_stats[k]['stats']][-1]
            if stats:
                try:
                    resources = stats['stats']['resources']
                    task_id = stats['task_id']
                except KeyError:
                    continue
                try:
                    filesize = sum(f['size'] for f in stats['stats']['task_stats']['upload'])
                    resources['filesize'] = filesize
                except KeyError:
                    resources['filesize'] = -1

                name = tasks[task_id]['name']
                job_index = jobs[tasks[task_id]['job_id']]['job_index']
                print(job_index, name, resources)
                dataset_stats[name][job_index] = resources
    except KeyboardInterrupt:
        print('stopping...')

    with open(args['output'], 'w') as f:
        json.dump(dataset_stats, f)

def main():
    parser = argparse.ArgumentParser(description='manually run IceProd i3exec')
    parser.add_argument('-t', '--token',help='auth token')
    parser.add_argument('-d','--dataset',type=int,help='dataset number')
    parser.add_argument('-o','--output',type=str,help='output json file')
    
    args = parser.parse_args()
    args = vars(args)

    logging.basicConfig(level=logging.WARNING)

    asyncio.run(get_stats(args))

if __name__ == '__main__':
    main()



"""
Test script for gridftp
"""

from __future__ import absolute_import, division, print_function

from tests.util import unittest_reporter, glob_tests

import logging
logger = logging.getLogger('gridftp')

import os
import shutil
import random
import subprocess
import tempfile

try:
    pass
except:
    pass

import unittest


import iceprod.core.gridftp

skip_tests = False
if (subprocess.call(['which','grid-proxy-init']) or
    subprocess.call(['grid-proxy-info','-e','-valid','1:0'])):
    skip_tests = True
    
skip_uberftp_tests = skip_tests
if (subprocess.call(['which','uberftp']) or
    subprocess.call(['which','globus-url-copy'])):
    skip_uberftp_tests = True


class gridftp_test(unittest.TestCase):
    def setUp(self):
        super(gridftp_test,self).setUp()

        self._timeout = 1
        self.test_dir = tempfile.mkdtemp(dir=os.getcwd())
        self.server_test_dir = os.path.join('gsiftp://gridftp.icecube.wisc.edu/data/sim/sim-new/tmp/test',
                                            str(random.randint(0,2**32)))

        if not skip_tests:
            try:
                iceprod.core.gridftp.GridFTP.mkdir(self.server_test_dir,
                                                   parents=True,
                                                   request_timeout=self._timeout)
            except:
                pass
            if not os.path.exists(self.test_dir):
                os.mkdir(self.test_dir)

            def cleanup():
                try:
                    iceprod.core.gridftp.GridFTP.rmtree(self.server_test_dir,
                                                        request_timeout=self._timeout)
                except:
                    pass
                shutil.rmtree(self.test_dir)
            self.addCleanup(cleanup)

    @unittest_reporter(skip=skip_tests)
    def test_01_supported_address(self):
        """Test supported_address"""
        bad_addresses = ['test','file:/test','gsiftp:test','gsiftp:/test',
                         'ftp:test','http://x2100.icecube.wisc.edu',
                         'ftp:/test']
        good_addresses = ['gsiftp://data.icecube.wisc.edu','ftp://gnu.org',
                          'gsiftp://gridftp-rr.icecube.wisc.edu/data/sim/sim-new']

        for i in range(0,10):
            for a in bad_addresses:
                ret = iceprod.core.gridftp.GridFTP.supported_address(a)
                if ret is True:
                    raise Exception('Bad address %s was called good'%a)
            for a in good_addresses:
                ret = iceprod.core.gridftp.GridFTP.supported_address(a)
                if ret is not True:
                    raise Exception('Good address %s was called bad'%a)

    @unittest_reporter(skip=skip_tests)
    def test_02_address_split(self):
        """Test address_split"""
        good_addresses = {'gsiftp://data.icecube.wisc.edu':('gsiftp://data.icecube.wisc.edu','/'),
                      'ftp://gnu.org':('ftp://gnu.org','/'),
                      'gsiftp://gridftp-rr.icecube.wisc.edu/data/sim/sim-new':('gsiftp://gridftp-rr.icecube.wisc.edu','/data/sim/sim-new')}

        for i in range(0,10):
            for a in good_addresses:
                pieces = iceprod.core.gridftp.GridFTP.address_split(a)
                if pieces != good_addresses[a]:
                    raise Exception('Address %s was not split properly'%a)

    @unittest_reporter(skip=skip_tests,name='put() with str')
    def test_100_put_str(self):
        """Test put with a str - synchronous"""
        address = os.path.join(self.server_test_dir,'test')
        filecontents = 'this is a test'

        iceprod.core.gridftp.GridFTP.put(address,data=filecontents,
                                         request_timeout=self._timeout)

    @unittest_reporter(skip=skip_tests,name='put() with file')
    def test_101_put_file(self):
        """Test put with a file - synchronous"""
        address = os.path.join(self.server_test_dir,'test')
        # make temp file
        filename = os.path.join(self.test_dir,'test')
        filecontents = 'this is a test'
        with open(filename,'w') as f:
            f.write(filecontents)

        iceprod.core.gridftp.GridFTP.put(address,filename=filename,
                                         request_timeout=self._timeout)

    @unittest_reporter(skip=skip_tests,name='get() with str')
    def test_110_get_str(self):
        """Test get with a str - synchronous"""
        address = os.path.join(self.server_test_dir,'test')
        filecontents = 'this is a test'

        # put str
        iceprod.core.gridftp.GridFTP.put(address,data=filecontents,
                                         request_timeout=self._timeout)

        # get str
        ret = iceprod.core.gridftp.GridFTP.get(address,
                                               request_timeout=self._timeout)
        self.assertEqual(ret, filecontents)

    @unittest_reporter(skip=skip_tests,name='get() with file')
    def test_111_get_file(self):
        """Test get with a file - synchronous"""
        address = os.path.join(self.server_test_dir,'test')
        # make temp file
        filename = os.path.join(self.test_dir,'test')
        filename2 = os.path.join(self.test_dir,'test2')
        filecontents = 'this is a test'
        with open(filename,'w') as f:
            f.write(filecontents)

        # put file
        iceprod.core.gridftp.GridFTP.put(address,filename=filename,
                                         request_timeout=self._timeout)

        # get file
        ret = iceprod.core.gridftp.GridFTP.get(address,filename=filename2,
                                               request_timeout=self._timeout)
        if not os.path.exists(filename2):
            raise Exception('dest file does not exist')
        newcontents = open(filename2).read()
        self.assertEqual(filecontents, newcontents)

    @unittest_reporter(skip=skip_tests,name='list(dir)')
    def test_120_list(self):
        """Test list of directory - synchronous"""
        address = os.path.join(self.server_test_dir,'test')

        iceprod.core.gridftp.GridFTP.mkdir(address,
                                           request_timeout=self._timeout)

        # get listing
        ret = iceprod.core.gridftp.GridFTP.list(address)
        self.assertEqual(ret, [])

    @unittest_reporter(skip=skip_tests,name='list(file)')
    def test_121_list(self):
        """Test list of file - synchronous"""
        address = os.path.join(self.server_test_dir,'test_file')
        data = 'this is a test'

        iceprod.core.gridftp.GridFTP.put(address,data=data,
                                               request_timeout=self._timeout)

        # get listing
        ret = iceprod.core.gridftp.GridFTP.list(address,
                                                request_timeout=self._timeout)
        self.assertEqual(ret, ['test_file'])

    @unittest_reporter(skip=skip_tests,name='list(dir,dotfiles)')
    def test_122_list(self):
        """Test list of dir with dotfiles - synchronous"""
        address = os.path.join(self.server_test_dir,'test')

        iceprod.core.gridftp.GridFTP.mkdir(address,
                                           request_timeout=self._timeout)


        # get listing
        ret = iceprod.core.gridftp.GridFTP.list(address,dotfiles=True,
                                                request_timeout=self._timeout)
        self.assertEqual(ret, ['.','..'])

    @unittest_reporter(skip=skip_tests,name='list(file,dotfiles)')
    def test_123_list(self):
        """Test list of file with dotfiles - synchronous"""
        address = os.path.join(self.server_test_dir,'test_file')
        data = 'this is a test'

        iceprod.core.gridftp.GridFTP.put(address,data=data,
                                         request_timeout=self._timeout)

        # get listing
        ret = iceprod.core.gridftp.GridFTP.list(address,dotfiles=True,
                                                request_timeout=self._timeout)
        self.assertEqual(ret, ['test_file'])

    @unittest_reporter(skip=skip_tests,name='list(dir,details)')
    def test_124_list(self):
        """Test list of dir with details - synchronous"""
        address = os.path.join(self.server_test_dir,'test')

        iceprod.core.gridftp.GridFTP.mkdir(address,
                                           request_timeout=self._timeout)

        # get listing
        ret = iceprod.core.gridftp.GridFTP.list(address,details=True,
                                                request_timeout=self._timeout)
        self.assertEqual(ret, [])

    @unittest_reporter(skip=skip_tests,name='list(file,details)')
    def test_125_list(self):
        """Test list of file with details - synchronous"""
        address = os.path.join(self.server_test_dir,'test_file')
        data = 'this is a test'

        iceprod.core.gridftp.GridFTP.put(address,data=data,
                                         request_timeout=self._timeout)

        # get listing
        ret = iceprod.core.gridftp.GridFTP.list(address,details=True,
                                                request_timeout=self._timeout)
        if len(ret) != 1 or ret[0].directory:
            logger.info('actual: %r',ret)
            raise Exception('list did not return expected results')

    @unittest_reporter(skip=skip_tests,name='list(dir,details,dotfiles)')
    def test_126_list(self):
        """Test list of dir with details and dotfiles - synchronous"""
        address = os.path.join(self.server_test_dir,'test')

        ret = iceprod.core.gridftp.GridFTP.mkdir(address,
                                                 request_timeout=self._timeout)

        # get listing
        ret = iceprod.core.gridftp.GridFTP.list(address,details=True,
                                                dotfiles=True,
                                                request_timeout=self._timeout)

        if (len(ret) != 2 or not any([x.name == '.' for x in ret])
            or not any([x.name == '..' for x in ret])):
            logger.info("expected: ['..','.']")
            logger.info('actual: %r',ret)
            raise Exception('list did not return expected results')

    @unittest_reporter(skip=skip_tests,name='list(file,details,dotfiles)')
    def test_127_list(self):
        """Test list of file with details and dotfiles - synchronous"""
        address = os.path.join(self.server_test_dir,'test_file')
        data = 'this is a test'

        iceprod.core.gridftp.GridFTP.put(address,data=data,
                                         request_timeout=self._timeout)

        # get listing
        ret = iceprod.core.gridftp.GridFTP.list(address,details=True,
                                                dotfiles=True,
                                                request_timeout=self._timeout)

        if len(ret) != 1 or ret[0].name != 'test_file':
            logger.info("expected: ['test_file']")
            logger.info('actual: %r',ret)
            raise Exception('list did not return expected results')

    @unittest_reporter(skip=skip_tests)
    def test_130_delete(self):
        """Test delete - synchronous"""
        address = os.path.join(self.server_test_dir,'test')
        filecontents = 'this is a test'

        # put str
        iceprod.core.gridftp.GridFTP.put(address,data=filecontents,
                                         request_timeout=self._timeout)

        iceprod.core.gridftp.GridFTP.delete(address,
                                            request_timeout=self._timeout)

    @unittest_reporter(skip=skip_uberftp_tests,name='rmtree(file)')
    def test_140_rmtree(self):
        """Test rmtree of a file - synchronous"""
        address = os.path.join(self.server_test_dir,'file_test')
        filecontents = 'this is a test'

        # put str
        iceprod.core.gridftp.GridFTP.put(address,data=filecontents,
                                         request_timeout=self._timeout)

        iceprod.core.gridftp.GridFTP.rmtree(address,
                                            request_timeout=self._timeout)

    @unittest_reporter(skip=skip_uberftp_tests,name='rmtree(empty dir)')
    def test_141_rmtree(self):
        """Test rmtree of an empty dir - synchronous"""
        address = os.path.join(self.server_test_dir,'test')

        # mkdir
        iceprod.core.gridftp.GridFTP.mkdir(address,
                                           request_timeout=self._timeout)

        iceprod.core.gridftp.GridFTP.rmtree(address,
                                            request_timeout=self._timeout)

    @unittest_reporter(skip=skip_uberftp_tests,name='rmtree(dir + file)')
    def test_142_rmtree(self):
        """Test rmtree of a directory with a file - synchronous"""
        address = os.path.join(self.server_test_dir,'test')

        # mkdir
        iceprod.core.gridftp.GridFTP.mkdir(address,
                                           request_timeout=self._timeout)

        address2 = os.path.join(self.server_test_dir,'test','file_test')
        filecontents = 'this is a test'

        # put str
        iceprod.core.gridftp.GridFTP.put(address2,data=filecontents,
                                         request_timeout=self._timeout)

        iceprod.core.gridftp.GridFTP.rmtree(address,
                                            request_timeout=self._timeout)

    @unittest_reporter(skip=skip_uberftp_tests,name='rmtree(dir + dir + file)')
    def test_143_rmtree(self):
        """Test rmtree of dir with subdir and subfile - synchronous"""
        address = os.path.join(self.server_test_dir,'test')

        # mkdir
        iceprod.core.gridftp.GridFTP.mkdir(address,
                                           request_timeout=self._timeout)

        # mkdir
        address2 = os.path.join(self.server_test_dir,'test','test2')
        iceprod.core.gridftp.GridFTP.mkdir(address2,
                                          request_timeout=self._timeout)

        address3 = os.path.join(self.server_test_dir,'test','test2','file_test')
        filecontents = 'this is a test'

        # put str
        iceprod.core.gridftp.GridFTP.put(address3,data=filecontents,
                                         request_timeout=self._timeout)

        iceprod.core.gridftp.GridFTP.rmtree(address,
                                            request_timeout=self._timeout)

    @unittest_reporter(skip=skip_tests)
    def test_160_exists(self):
        """Test exists - synchronous"""
        address = os.path.join(self.server_test_dir,'test')
        filecontents = 'this is a test'

        ret = iceprod.core.gridftp.GridFTP.exists(address,
                                                  request_timeout=self._timeout)
        if ret is True:
            raise Exception('exists succeeded when it should have failed')

        # put str
        iceprod.core.gridftp.GridFTP.put(address,data=filecontents,
                                         request_timeout=self._timeout)

        ret = iceprod.core.gridftp.GridFTP.exists(address,
                                                  request_timeout=self._timeout)
        if ret is not True:
            raise Exception('exists failed: ret=%r'%ret)

    @unittest_reporter(skip=skip_tests)
    def test_170_move(self):
        """Test move - synchronous"""
        address = os.path.join(self.server_test_dir,'test')
        address2 = os.path.join(self.server_test_dir,'test2')
        filecontents = 'this is a test'

        # put str
        iceprod.core.gridftp.GridFTP.put(address,data=filecontents,
                                         request_timeout=self._timeout)

        ret = iceprod.core.gridftp.GridFTP.exists(address2,
                                                  request_timeout=self._timeout)
        if ret is True:
            raise Exception('exists succeeded before move')

        iceprod.core.gridftp.GridFTP.move(address,address2,
                                          request_timeout=self._timeout)
        
        ret = iceprod.core.gridftp.GridFTP.exists(address,
                                                  request_timeout=self._timeout)
        if ret is True:
            raise Exception('exists succeeded on old address')
        ret = iceprod.core.gridftp.GridFTP.exists(address2,
                                                  request_timeout=self._timeout)
        if ret is not True:
            raise Exception('exists failed on new address')

    @unittest_reporter(skip=skip_uberftp_tests)
    def test_180_checksum(self):
        """Test checksums - synchronous"""
        address = os.path.join(self.server_test_dir,'test')
        filecontents = b'this is a test'

        import hashlib

        # put str
        iceprod.core.gridftp.GridFTP.put(address,data=filecontents,
                                         request_timeout=self._timeout)

        ret = iceprod.core.gridftp.GridFTP.md5sum(address,
                                                  request_timeout=self._timeout)
        correct = hashlib.md5(filecontents).hexdigest()
        if ret != correct:
            raise Exception('md5sum failed: ret=%r and correct=%r'%(ret,correct))

        ret = iceprod.core.gridftp.GridFTP.sha1sum(address,
                                                   request_timeout=self._timeout)
        correct = hashlib.sha1(filecontents).hexdigest()
        if ret != correct:
            raise Exception('sha1sum failed: ret=%r and correct=%r'%(ret,correct))

        ret = iceprod.core.gridftp.GridFTP.sha256sum(address,
                                                     request_timeout=self._timeout)
        correct = hashlib.sha256(filecontents).hexdigest()
        if ret != correct:
            raise Exception('sha256sum failed: ret=%r and correct=%r'%(ret,correct))

        ret = iceprod.core.gridftp.GridFTP.sha512sum(address,
                                                     request_timeout=self._timeout)
        correct = hashlib.sha512(filecontents).hexdigest()
        if ret != correct:
            raise Exception('sha512sum failed: ret=%r and correct=%r'%(ret,correct))

    @unittest_reporter(skip=skip_uberftp_tests)
    def test_190_size(self):
        """Test size - synchronous"""
        address = os.path.join(self.server_test_dir,'test')
        filecontents = 'this is a test'

        # put str
        iceprod.core.gridftp.GridFTP.put(address,data=filecontents,
                                         request_timeout=self._timeout)

        ret = iceprod.core.gridftp.GridFTP.size(address,
                                                request_timeout=self._timeout)
        correct = len(filecontents)
        self.assertEqual(ret, correct)


def load_tests(loader, tests, pattern):
    suite = unittest.TestSuite()
    alltests = glob_tests(loader.getTestCaseNames(gridftp_test))
    suite.addTests(loader.loadTestsFromNames(alltests,gridftp_test))
    return suite



"""
Clean job temp directories.

Check job temp directories, and if the job is complete then delete it.
"""

import argparse
import asyncio
from collections import defaultdict
from datetime import datetime, timedelta, UTC
import logging
import os
import shutil
from typing import Any
from wipac_dev_tools import from_environment

from iceprod.client_auth import add_auth_to_argparse, create_rest_client
from iceprod.server.util import str2datetime

logger = logging.getLogger('job_temp_cleaning')


listing = dict[str, list[str]]


def list_dataset_job_dirs_fs(path: str, *, prefix: str | None = None) -> listing:
    dataset_dirs = defaultdict(list)
    if prefix:
        path = os.path.join(path, prefix)
        dataset_dirs[prefix] = os.listdir(path)
    else:
        dirs = os.listdir(path)
        for d in dirs:
            p = os.path.join(path, d)
            if os.path.isdir(p):
                dataset_dirs[d] = os.listdir(p)
    return dataset_dirs


def filter_job_dirs(job_dirs: list[str], job_indexes: set[str], debug: bool = False):
    for j in job_dirs:
        if not j.isnumeric():
            if debug:
                logger.info('j is not numeric: %r', j)
                raise Exception('not numeric')
            continue
        if int(j) in job_indexes:
            yield j


async def clean_dataset_dir(dataset_num: str, dataset_id: str, job_dirs: list[str], temp_dir: str, rmtree: Any, debug: bool, rest_client: Any):
    logger.info('temp cleaning for dataset %r', dataset_num)
    logger.debug('job_dirs: %r', job_dirs)

    now = datetime.now(UTC)
    suspend_time = timedelta(days=90)

    try:
        jobs = await rest_client.request('GET', f'/datasets/{dataset_id}/jobs', {'keys':'status|status_changed|job_index'})
    except Exception:
        logger.error('failed to get jobs for dataset %r', dataset_num, exc_info=True)
        if debug:
            raise
        return
    logger.debug('jobs: %r', jobs)

    job_indexes = set()
    for job in jobs.values():
        if job['status'] == 'complete' or (job['status'] != 'processing' and now - str2datetime(job['status_changed']) > suspend_time):
            job_indexes.add(job['job_index'])
    logger.debug('job_indexes: %r', job_indexes)

    futures = set()  # type: ignore
    for j in filter_job_dirs(job_dirs, job_indexes, debug=debug):
        while len(futures) >= 16:
            done, futures = await asyncio.wait(futures, return_when=asyncio.FIRST_COMPLETED)
            for f in done:
                try:
                    await f
                except Exception:
                    logger.warning('failed to clean site_temp', exc_info=True)
                    if debug:
                        raise

        dagtemp = os.path.join(temp_dir, dataset_num, j)
        logger.info('cleaning site_temp %r', dagtemp)
        ret = rmtree(dagtemp)
        if ret and asyncio.iscoroutine(ret):
            futures.add(asyncio.create_task(ret))

    if futures:
        for f in asyncio.as_completed(futures):
            await f


async def run(rest_client, temp_dir, list_dirs, rmtree, dataset=None, debug=False):
    """
    Actual runtime / loop.

    Args:
        rest_client (:py:class:`iceprod.core.rest_client.Client`): rest client
        temp_dir (str): temporary directory to clean
        list_dirs (callable): list temporary directory
        rmtree (callable): delete temporary directory and contents
        dataset (int): dataset num to run over (optional)
        debug (bool): debug flag to propagate exceptions
    """
    try:
        # get all the job_indexes currently in tmp
        if dataset:
            dataset_dirs = list_dirs(temp_dir, prefix=str(dataset))
            if dataset_dirs and asyncio.iscoroutine(dataset_dirs):
                dataset_dirs = await dataset_dirs
        else:
            dataset_dirs = list_dirs(temp_dir)
            if dataset_dirs and asyncio.iscoroutine(dataset_dirs):
                dataset_dirs = await dataset_dirs
        logger.debug('dataset_dirs: %r', dataset_dirs)

        ret = await rest_client.request('GET', '/datasets?keys=dataset_id|dataset')
        logger.debug('datasets api raw: %r', ret)
        datasets = {}
        for d in ret:
            datasets[str(ret[d]['dataset'])] = ret[d]['dataset_id']
        logger.debug('datasets: %r', datasets)

        for d in dataset_dirs:
            await clean_dataset_dir(d, datasets[d], dataset_dirs[d], temp_dir, rmtree, debug=debug, rest_client=rest_client)

    except Exception:
        logger.error('error checking job temp', exc_info=True)
        if debug:
            raise


def main():
    default_config = {
        'SITE_TEMP': '/scratch',
    }
    config = from_environment(default_config)

    parser = argparse.ArgumentParser(description='job temp cleaning')
    add_auth_to_argparse(parser)
    parser.add_argument('-d', '--dataset', type=str, help='dataset num (optional)')
    parser.add_argument('--site-temp', default=config['SITE_TEMP'], help='site temp location')
    parser.add_argument('--log-level', default='info', help='log level')
    parser.add_argument('--debug', default=False, action='store_true', help='debug enabled')

    args = parser.parse_args()

    logformat = '%(asctime)s %(levelname)s %(name)s %(module)s:%(lineno)s - %(message)s'
    logging.basicConfig(format=logformat, level=getattr(logging, args.log_level.upper()))

    rest_client = create_rest_client(args)

    if os.path.exists(args.site_temp):
        logging.info('using local filesystem')
        listdir = list_dataset_job_dirs_fs
        rmtree = shutil.rmtree
    else:
        raise RuntimeError('unknown type of scratch')

    logging.info('temp dir: %r', args.site_temp)
    asyncio.run(run(rest_client, args.site_temp, list_dirs=listdir, rmtree=rmtree, dataset=args.dataset, debug=args.debug))


if __name__ == '__main__':
    main()



import logging
import pytest
from rest_tools.client import RestClient

from iceprod.core.config import ConfigSchema, Config, Dataset, Job, Task
from iceprod.server.util import nowstr


def test_config_schema():
    assert ConfigSchema.list_versions() == [3.1, 3.2]

    assert ConfigSchema.schema()['title'] == 'IceProd Dataset Config'
    assert ConfigSchema.schema(3.2)['properties']['version']['default'] == 3.2

    assert ConfigSchema.data_defaults(3.2)['remote'] == ''

def test_config():
    config = {'tasks':[
        {
            'name': 'testing',
            'trays': [{
                'modules': [{
                    'src': '/usr/bin/python3',
                    'args': ''
                }]
            }],
            'data': [
                {
                    'remote': 'token:///data/sim/IceCube/2025/file.i3.zst',
                    'movement': 'input'
                }
            ]
        }
    ]}
    c = Config(config)
    c.fill_defaults()
    c.validate()


def test_config_str_ver():
    config = {
        'version': '3.2',
        'tasks':[{
            'name': 'testing',
            'trays': [{
                'modules': [{
                    'src': '/usr/bin/python3',
                    'args': ''
                }]
            }]
        }
    ]}
    c = Config(config)
    c.fill_defaults()
    c.validate()


def test_config_invalid_data():
    config = {
        'version': 3.2,
        'tasks':[{
            'name': 'testing',
            'trays': [{
                'modules': [{
                    'src': '/usr/bin/python3',
                    'args': ''
                }]
            }],
            'data': [
                {
                    'type': 'job_temp',
                    'remote': 'token:///data/sim/IceCube/2025/file.i3.zst',
                    'movement': 'input'
                }
            ]
        }
    ]}
    c = Config(config)
    c.fill_defaults()
    with pytest.raises(Exception, match="remote should be empty"):
        c.validate()


def test_config_manual_scope():
    config = {
        'version': 3.2,
        'tasks':[{
            'name': 'testing',
            'trays': [{
                'modules': [{
                    'src': '/usr/bin/python3',
                    'args': ''
                }]
            }],
            'task_files': True,
        }
    ]}
    c = Config(config)
    c.fill_defaults()
    with pytest.raises(Exception, match="token_scopes are required"):
        c.validate()


def test_config_dups():
    config = {'tasks':[
        {
            'name': 'testing',
            'trays': [{
                'modules': [{
                    'src': '/usr/bin/python3',
                    'args': ''
                }]
            }],
            'data': [
                {
                    'remote': 'token:///data/sim/IceCube/2025/file.i3.zst',
                    'movement': 'input'
                }
            ]
        },
        {
            'name': 'testing2',
            'trays': [{
                'modules': [{
                    'src': '/usr/bin/python3',
                    'args': ''
                }]
            }],
            'data': [
                {
                    'remote': 'token:///data/sim/IceCube/2025/file.i3.zst',
                    'movement': 'input'
                }
            ]
        }
    ]}
    c = Config(config)
    c.fill_defaults()
    c.validate()

    c.config['tasks'][0]['token_scopes']['foo'] = 'bar'
    assert 'foo' not in c.config['tasks'][1]['token_scopes']


def test_dataset_dataclasses():
    with pytest.raises(Exception):
        Dataset()

    d = Dataset('did123', 123, 1, 2, 3, 'processing', 0.5, 'grp', 'usr', False, config={})
    assert d.dataset_id == 'did123'
    assert d.dataset_num == 123
    assert d.jobs_submitted == 1
    assert d.tasks_submitted == 2
    assert d.tasks_per_job == 3
    assert d.status == 'processing'
    assert d.priority == 0.5
    assert d.group == 'grp'
    assert d.user == 'usr'
    assert d.debug is False
    assert d.config == {}


async def test_load_config(requests_mock):
    dataset_id = 'did123'
    dataset_data = {
        'dataset': 123,
        'dataset_id': 'did123',
        'status': 'processing',
        'jobs_submitted': 1,
        'tasks_submitted': 1,
        'tasks_per_job': 1,
        'priority': 0.5,
        'group': 'g123',
        'username': 'u123',
        'debug': False
    }
    requests_mock.get(f'http://test.iceprod/datasets/{dataset_id}', json=dataset_data)
    config_data = {
        'my': 'config'
    }
    requests_mock.get(f'http://test.iceprod/config/{dataset_id}', json=config_data)

    r = RestClient('http://test.iceprod')
    d = await Dataset.load_from_api(dataset_id, r)

    
    assert d.dataset_id == dataset_id
    assert d.dataset_num == dataset_data['dataset']
    assert d.group == dataset_data['group']
    assert d.user == dataset_data['username']
    assert d.config == config_data


async def test_defaults():
    dataset_data = {
        'dataset': 123,
        'dataset_id': 'did123',
        'status': 'processing',
        'jobs_submitted': 1,
        'tasks_submitted': 1,
        'tasks_per_job': 1,
        'priority': 0.5,
        'group': 'g123',
        'username': 'u123',
        'debug': False
    }
    config_data = {}
    d = Dataset(
        dataset_data['dataset_id'],
        dataset_data['dataset'],
        dataset_data['jobs_submitted'],
        dataset_data['tasks_submitted'],
        dataset_data['tasks_per_job'],
        dataset_data['status'],
        dataset_data['priority'],
        dataset_data['group'],
        dataset_data['username'],
        dataset_data['debug'],
        config_data
    )
    d.fill_defaults()
    logging.info('after defaults: %r', d.config)
    assert d.config['version'] == 3.1
    assert d.config['options'] == {}
    assert d.config['steering'] == {'parameters': {}, 'batchsys': {}, 'data': []}

    
async def test_defaults_refs():
    dataset_data = {
        'dataset': 123,
        'dataset_id': 'did123',
        'status': 'processing',
        'jobs_submitted': 1,
        'tasks_submitted': 1,
        'tasks_per_job': 1,
        'priority': 0.5,
        'group': 'g123',
        'username': 'u123',
        'debug': False
    }
    config_data = {
        'steering': {
            'parameters': {'a': 'b'}
        },
        'tasks': [{
            'requirements': {},
            'data': [{
                'remote': 'http://test/file'
            }],
            'trays': [{
                'modules': [{}]
            }]
        }]
    }
    d = Dataset(
        dataset_data['dataset_id'],
        dataset_data['dataset'],
        dataset_data['jobs_submitted'],
        dataset_data['tasks_submitted'],
        dataset_data['tasks_per_job'],
        dataset_data['status'],
        dataset_data['priority'],
        dataset_data['group'],
        dataset_data['username'],
        dataset_data['debug'],
        config_data
    )
    d.fill_defaults()
    logging.info('after defaults: %r', d.config)
    assert d.config['tasks'][0]['requirements']['cpu'] == 1
    assert d.config['tasks'][0]['requirements']['memory'] == 1.0
    assert d.config['tasks'][0]['requirements']['disk'] == 1.0
    assert d.config['tasks'][0]['requirements']['time'] == 1.0
    assert d.config['tasks'][0]['data'][0]['local'] == ''
    assert d.config['tasks'][0]['data'][0]['type'] == 'permanent'
    assert d.config['tasks'][0]['data'][0]['movement'] == 'input'
    assert d.config['tasks'][0]['data'][0]['transfer'] is True
    assert d.config['tasks'][0]['trays'][0]['iterations'] == 1
    assert d.config['tasks'][0]['trays'][0]['modules'][0]['env_shell'] == ''
    assert d.config['tasks'][0]['trays'][0]['modules'][0]['env_clear'] is True

async def test_validate_error():
    dataset_data = {
        'dataset': 123,
        'dataset_id': 'did123',
        'status': 'processing',
        'jobs_submitted': 1,
        'tasks_submitted': 1,
        'tasks_per_job': 1,
        'priority': 0.5,
        'group': 'g123',
        'username': 'u123',
        'debug': False
    }
    config_data = {
        'my': 'config'
    }
    d = Dataset(
        dataset_data['dataset_id'],
        dataset_data['dataset'],
        dataset_data['jobs_submitted'],
        dataset_data['tasks_submitted'],
        dataset_data['tasks_per_job'],
        dataset_data['status'],
        dataset_data['priority'],
        dataset_data['group'],
        dataset_data['username'],
        dataset_data['debug'],
        config_data
    )
    with pytest.raises(Exception):
        d.validate()


async def test_validate_valid():
    dataset_data = {
        'dataset': 123,
        'dataset_id': 'did123',
        'status': 'processing',
        'jobs_submitted': 1,
        'tasks_submitted': 1,
        'tasks_per_job': 1,
        'priority': 0.5,
        'group': 'g123',
        'username': 'u123',
        'debug': False
    }
    config_data = {
        'tasks': [{
            'name': 'first',
            'trays': [{
                'modules': [{}]
            }]
        }]
    }
    d = Dataset(
        dataset_data['dataset_id'],
        dataset_data['dataset'],
        dataset_data['jobs_submitted'],
        dataset_data['tasks_submitted'],
        dataset_data['tasks_per_job'],
        dataset_data['status'],
        dataset_data['priority'],
        dataset_data['group'],
        dataset_data['username'],
        dataset_data['debug'],
        config_data
    )
    d.fill_defaults()
    d.validate()


def test_job_dataclasses():
    with pytest.raises(Exception):
        Job()

    d = Dataset('did123', 123, 2, 1, 1, 'processing', 0.5, 'grp', 'usr', False, {})
    j = Job(d, 'j123', 1, 'processing')

    assert j.dataset == d
    assert j.job_id == 'j123'
    assert j.job_index == 1
    assert j.status == 'processing'


def test_task_dataclasses():
    with pytest.raises(Exception):
        Task()

    d = Dataset('did123', 123, 2, 1, 1, 'processing', 0.5, 'grp', 'usr', False, {})
    j = Job(d, 'j123', 1, 'processing')
    t = Task(d, j, 't123', 0, 'foo', [], {}, 'waiting', '', {})

    assert t.dataset == d
    assert t.job == j
    assert t.task_id == 't123'
    assert t.task_index == 0
    assert t.name == 'foo'
    assert t.depends == []
    assert t.requirements == {}
    assert t.status == 'waiting'
    assert t.site == ''
    assert t.stats == {}
    assert not t.task_files
    assert not t.instance_id


def test_task_config():
    d = Dataset('did123', 123, 2, 1, 1, 'processing', 0.5, 'grp', 'usr', False, {'tasks':[1,2,3]})
    j = Job(d, 'j123', 1, 'processing')
    t = Task(d, j, 't123', 0, 'foo', [], {}, 'waiting', '', {})

    assert t.get_task_config() == 1


async def test_task_load_from_api(requests_mock):
    dataset_id = 'did123'
    dataset_data = {
        'dataset': 123,
        'dataset_id': 'did123',
        'status': 'processing',
        'jobs_submitted': 1,
        'tasks_submitted': 1,
        'tasks_per_job': 1,
        'priority': 0.5,
        'group': 'g123',
        'username': 'u123',
        'debug': False
    }
    requests_mock.get(f'http://test.iceprod/datasets/{dataset_id}', json=dataset_data)
    config_data = {
        'my': 'config'
    }
    requests_mock.get(f'http://test.iceprod/config/{dataset_id}', json=config_data)
    job_data = {
        'dataset_id': dataset_id,
        'job_id': 'j123',
        'job_index': 1,
        'status': 'processing',
    }
    requests_mock.get(f'http://test.iceprod/datasets/{dataset_id}/jobs/{job_data["job_id"]}', json=job_data)
    task_data = job_data | {
        'task_id': 't123',
        'task_index': 0,
        'name': 'foo',
        'depends': [],
        'requirements': {'cpu': 1},
        'status': 'waiting',
        'status_changed': nowstr(),
        'failures': 1,
        'evictions': 0,
        'walltime': 0.0,
        'walltime_err': 0.15,
        'walltime_err_n': 1,
        'site': 'CHTC',
    }
    requests_mock.get(f'http://test.iceprod/datasets/{dataset_id}/tasks/{task_data["task_id"]}', json=task_data)

    r = RestClient('http://test.iceprod')
    t = await Task.load_from_api(dataset_id, task_data['task_id'], r)

    assert t.dataset.dataset_id == dataset_id
    assert t.job.job_id == job_data['job_id']
    assert t.task_id == task_data['task_id']


async def test_task_load_stats(requests_mock):
    config = {'tasks':[
        {
            'name': 'testing',
            'trays': [{
                'modules': [{
                    'src': '/usr/bin/python3',
                    'args': ''
                }]
            }]
        }
    ]}
    d = Dataset('did123', 123, 2, 1, 1, 'processing', 0.5, 'grp', 'usr', False, config)
    d.fill_defaults()
    d.validate()
    j = Job(d, 'j123', 1, 'processing')
    t = Task(d, j, 't123', 0, 'foo', [], {}, 'waiting', '', {})

    stat_data = {
        'task_stat_id': 'ts123',
        'foo': 'bar',
    }
    requests_mock.get(
        'http://test.iceprod/datasets/did123/tasks/t123/task_stats',
        json={stat_data['task_stat_id']: stat_data},
    )

    r = RestClient('http://test.iceprod')
    await t.load_stats_from_api(r)

    assert t.stats == stat_data


async def test_task_load_task_files(requests_mock):
    config = {'tasks':[
        {
            'name': 'testing',
            'trays': [{
                'modules': [{
                    'src': '/usr/bin/python3',
                    'args': ''
                }]
            }],
            'task_files': True
        }
    ]}
    d = Dataset('did123', 123, 2, 1, 1, 'processing', 0.5, 'grp', 'usr', False, config)
    d.fill_defaults()
    d.validate()
    j = Job(d, 'j123', 1, 'processing')
    t = Task(d, j, 't123', 0, 'foo', [], {}, 'waiting', '', {})

    data = [{
        'local': 'foo.txt',
        'remote': 'https://foo.bar.baz/foo.txt',
    }]
    requests_mock.get(
        'http://test.iceprod/datasets/did123/files/t123',
        json={'files': data},
    )

    r = RestClient('http://test.iceprod')
    await t.load_task_files_from_api(r)

    assert t.task_files
    assert len(t.task_files) == len(data)
    assert t.task_files[0]['local'] == data[0]['local']
    assert t.task_files[0]['remote'] == data[0]['remote']
    assert t.task_files[0]['type'] == 'permanent'
    assert t.task_files[0]['movement'] == 'input'



import asyncio
from contextlib import asynccontextmanager
from dataclasses import dataclass, asdict, KW_ONLY
from datetime import datetime, timedelta, timezone
import logging
from typing import Any, AsyncIterator, Literal
import uuid

from pymongo import ReturnDocument, ASCENDING, DESCENDING

from .mongo import Mongo, CollectionIndexes


type Payload = dict[str, Any]


class FailMessage(Exception):
    pass


@dataclass
class Message:
    """Mongo Queue Message"""
    payload: Payload
    _: KW_ONLY
    uuid: str
    status: Literal['queued', 'processing', 'error', 'complete']
    priority: int
    created_at: datetime
    attempts: int = 0
    locked_at: None | datetime = None
    worker_id: None | str = None
    error_message: None | str = None


class AsyncMongoQueue:
    """
    A basic queue system using MongoDB.

    Will retry after a timeout, and can rank messages by priority (lower is better).

    Args:
        url: mongo url
        collection_name: mongo collection
        worker_id: unique id for consumer / worker
        deadline: number of days messages should remain on the queue in any state
        extra_indexes: payload indexes to apply
        **mongo_args: any extra args to Mongo, such as write concern or timeout
    """
    def __init__(
        self,
        *,
        url: str,
        collection_name: str,
        worker_id: str | None = None,
        deadline: float = 1.0,
        extra_indexes: CollectionIndexes | None = None,
        **mongo_args
    ):
        self.client = Mongo(url=url, **mongo_args)
        self.collection_name = collection_name
        self.collection = self.client.db[collection_name]
        self.worker_id = worker_id or uuid.uuid4().hex
        self.deadline = deadline
        self.extra_indexes = extra_indexes

    async def close(self):
        await self.client.close()

    async def setup(self):
        """Initializes indexes for performance and priority fetching."""
        await self.client.ping()
        indexes = {
            self.collection_name: {
                'uuid_index_1': {
                    'keys': [
                        ('uuid', ASCENDING)
                    ],
                    'unique': True,
                },
                'queue_index_1': {
                    'keys': [
                        ("status", ASCENDING),
                        ("priority", DESCENDING),
                        ("attempts", ASCENDING),
                        ("created_at", ASCENDING)
                    ],
                }
            }
        }
        if self.extra_indexes:
            indexes[self.collection_name].update(self.extra_indexes)
        await self.client.create_indexes(indexes=indexes)
        asyncio.create_task(self.clean_task())

    async def push(self, payload: Payload, priority: int = 0) -> str:
        """
        Adds a new message to the queue.

        Args:
            payload: message data
            priority: priority of the message (higher is better)

        Returns:
            Message id
        """
        message = Message(
            uuid=uuid.uuid4().hex,
            payload=payload,
            status='queued',
            priority=priority,
            created_at=datetime.now(timezone.utc),
        )
        await self.collection.insert_one(asdict(message))
        return message.uuid

    async def push_if_not_exists(self, payload: Payload, filter_payload: Payload | None = None, priority: int = 0) -> str:
        """
        Adds a new message to the queue, deduplicating on payload.

        Args:
            payload: message data
            filter_payload: different payload to filter on for deduplication (if None, uses payload)
            priority: priority of the message (higher is better)

        Returns:
            Message id
        """
        message = Message(
            uuid=uuid.uuid4().hex,
            payload=payload,
            status='queued',
            priority=priority,
            created_at=datetime.now(timezone.utc),
        )
        payload_lookup = filter_payload if filter_payload is not None else payload
        query = {
            f'payload.{name}': value for name,value in payload_lookup.items()
        }
        query['status'] = {'$in': ['queued', 'processing']}
        update = {
            '$setOnInsert': asdict(message),
        }
        ret = await self.collection.find_one_and_update(
            filter=query,
            update=update,
            projection={'_id': False},
            return_document=ReturnDocument.AFTER,
            upsert=True
        )
        if not ret:
            raise Exception('failed to push')
        return ret['uuid']

    async def get_status(self, message_id: str) -> None | str:
        """Get the status if a message exists"""
        ret = await self.collection.find_one(
            {'uuid': message_id},
            projection={'_id': False, 'status': True}
        )
        if ret:
            return ret['status']
        else:
            return None

    async def get_error(self, message_id: str) -> None | str:
        """Get the error message if a message exists"""
        ret = await self.collection.find_one(
            {'uuid': message_id},
            projection={'_id': False, 'status': True, 'error_message': True}
        )
        if ret and ret['status'] == 'error':
            return ret.get('error_message', None)
        else:
            return None

    async def get_payload(self, message_id: str) -> None | Payload:
        ret = await self.collection.find_one(
            {'uuid': message_id},
            projection={'_id': False}
        )
        if ret:
            return ret['payload']
        else:
            return None

    async def lookup_by_payload(self, payload_lookup: Payload, **extra_args) -> Message | None:
        """
        Lookup a message by payload details.

        Must use mongo `.` notation for nested keys.
        """
        query = {
            f'payload.{name}': value for name,value in payload_lookup.items()
        }
        logging.info('payload_lookup query = %r', query)
        ret = await self.collection.find_one(
            query,
            projection={'_id': False},
            **extra_args
        )
        if ret:
            ret = Message(**ret)
        return ret

    async def count(self, query: dict[str, Any] | None = None) -> int:
        """Count how many requests are in the queue."""
        if not query:
            query = {}
        ret = await self.collection.count_documents(query, maxTimeMS=1000)
        return ret

    async def pop(self, timeout_seconds: float = 300.) -> Message | None:
        """Atomically grabs the next available message (or an abandoned timed-out one)."""
        timeout_cutoff = datetime.now(timezone.utc) - timedelta(seconds=timeout_seconds)

        query = {
            '$or': [
                {'status': 'queued'},
                {'status': 'processing', 'locked_at': {'$lt': timeout_cutoff}}
            ]
        }

        update = {
            '$set': {
                'status': 'processing',
                'locked_at': datetime.now(timezone.utc),
                'worker_id': self.worker_id
            },
            '$inc': {'attempts': 1},
        }

        # Atomically find and update
        ret = await self.collection.find_one_and_update(
            query,
            update,
            sort=[('priority', DESCENDING), ('attempts', ASCENDING), ('created_at', ASCENDING)],
            projection={'_id': False},
            return_document=ReturnDocument.AFTER
        )
        if ret:
            ret = Message(**ret)
        return ret

    async def update_payload(self, message_id: str, payload: Payload):
        """Update a payload"""
        update = {
            f'payload.{name}': value for name,value in payload.items()
        }
        await self.collection.update_one(
            {'uuid': message_id},
            {'$set': update}
        )

    async def complete(self, message_id: str):
        """Acknowledges and sets the state to 'completed' upon successful processing."""
        await self.collection.update_one(
            {'uuid': message_id},
            {'$set': {'status': 'complete'}}
        )

    async def release(self, message_id: str):
        """Re-releases the message back to 'queued' state."""
        await self.collection.update_one(
            {'uuid': message_id},
            {'$set': {'status': 'queued', 'locked_at': None, 'worker_id': None}}
        )

    async def fail(self, message_id: str, error_message: str | None = None):
        """Acknowledges and sets the state to 'error' upon a failed processing."""
        data = {'status': 'error'}
        if error_message:
            data['error_message'] = error_message
        await self.collection.update_one(
            {'uuid': message_id},
            {'$set': data}
        )

    @asynccontextmanager
    async def process_next(self, timeout_seconds=300) -> AsyncIterator[Payload | None]:
        """
        Context manager that pops a message and auto-handles completion/failure.
        """
        message = await self.pop(timeout_seconds=timeout_seconds)

        if not message:
            yield None
            return

        try:
            yield message.payload
            await self.complete(message.uuid)
        except Exception as e:
            await self.fail(message_id=message.uuid, error_message=str(e))
            raise

    async def clean(self):
        """Clean out old messages"""
        await self.collection.delete_many({
            'created_at': {'$lt': datetime.now(timezone.utc) - timedelta(days=self.deadline)},
        })

    async def clean_task(self):
        """Periodically run the clean function."""
        while True:
            try:
                await self.clean()
            except Exception:
                logging.info('mongo_queue.clean failed', exc_info=True)

            # make sure to run at least 10 times more often than the deadline interval
            await asyncio.sleep(self.deadline*8640)



"""
The queue module is responsible for interacting with the local batch or
queueing system, putting tasks on the queue and removing them as necessary.
"""
import asyncio
import importlib
import logging
from pathlib import Path
import pkgutil
import time

from rest_tools.client import ClientCredentialsAuth

from iceprod.server.globus import SiteGlobusProxy


class StopException(Exception):
    pass


logger = logging.getLogger('queue')


class Queue:
    """
    Run the queue module, which queues jobs onto the local grid system(s).
    """

    def __init__(self, cfg):
        self.cfg = cfg
        self.proxy = None
        self.proxy_task = None

        self.max_duration = 3600*12

        self.grid = self.get_grid_plugin()

    async def run(self):
        # set up x509 proxy
        proxy_kwargs = {'duration': self.max_duration}
        if 'gridftp_cfgfile' in self.cfg['queue']:
            proxy_kwargs['cfgfile'] = self.cfg['queue']['gridftp_cfgfile']
        self.proxy = SiteGlobusProxy(**proxy_kwargs)

        self.proxy_task = asyncio.create_task(self.check_proxy())

        try:
            await self.grid.run()
        finally:
            if self.proxy_task:
                self.proxy_task.cancel()
                self.proxy_task = None
            await asyncio.sleep(0)

    def get_grid_plugin(self):
        """
        Get the grid plugin called for in the config.

        Returns:
            Grid plugin instantiated class
        """
        plugin_desc = self.cfg['queue']['description']
        plugin_type = self.cfg['queue']['type']
        logger.info('queueing plugin in cfg: %r %r', plugin_desc, plugin_type)

        # try to find installed plugins
        plugin_path = str(Path(__file__).parent / 'plugins')
        raw_types = [name for _, name, _ in pkgutil.iter_modules([plugin_path])]

        logger.info('installed plugins: %r', raw_types)
        plugin_type = plugin_type.lower()
        p = None
        for raw_name in raw_types:
            if raw_name == plugin_type:
                # exact match
                logger.debug('exact plugin match - %s', raw_name)
                p = raw_name
                break
            elif plugin_type.startswith(raw_name):
                # partial match
                if p is None:
                    logger.debug('partial plugin match - %s', raw_name)
                    p = raw_name
                else:
                    if len(raw_name) > len(p):
                        logger.debug('better plugin match - %s', raw_name)
                        p = raw_name
        if p is not None:
            plugin_module = importlib.import_module(f'iceprod.server.plugins.{p}')
        else:
            raise RuntimeError(f'Cannot find plugin for type {plugin_type}')

        # get duration
        duration = 0
        if 'max_task_queued_time' in self.cfg['queue']:
            duration += self.cfg['queue']['max_task_queued_time']
        if 'max_task_processing_time' in self.cfg['queue']:
            duration += self.cfg['queue']['max_task_processing_time']
        if duration > self.max_duration:
            self.max_duration = duration

        # instantiate plugin
        kwargs = {
            'cfg': self.cfg,
            'rest_client': None,
            'cred_client': None,
        }

        if ('rest_api' in self.cfg and 'url' in self.cfg['rest_api']
                and 'oauth_url' in self.cfg['rest_api']
                and 'oauth_client_id' in self.cfg['rest_api']
                and 'oauth_client_secret' in self.cfg['rest_api']):
            try:
                kwargs['rest_client'] = ClientCredentialsAuth(
                    address=self.cfg['rest_api']['url'],
                    token_url=self.cfg['rest_api']['oauth_url'],
                    client_id=self.cfg['rest_api']['oauth_client_id'],
                    client_secret=self.cfg['rest_api']['oauth_client_secret'],
                )
                kwargs['cred_client'] = ClientCredentialsAuth(
                    address=self.cfg['rest_api']['cred_url'],
                    token_url=self.cfg['rest_api']['oauth_url'],
                    client_id=self.cfg['rest_api']['oauth_client_id'],
                    client_secret=self.cfg['rest_api']['oauth_client_secret'],
                    retries=1,
                )
            except Exception:
                logger.warning('failed to connect to rest api: %r',
                               self.cfg['rest_api'].get('url',''), exc_info=True)

        try:
            return getattr(plugin_module, 'Grid')(**kwargs)
        except Exception:
            logger.error('Error calling plugin class', exc_info=True)
            raise RuntimeError('Error calling plugin class')

    async def check_proxy(self, duration=None):
        """
        Check the x509 proxy every 5 minutes.
        """
        def blocking():
            try:
                assert self.proxy
                if duration:
                    self.proxy.set_duration(duration//3600)
                self.proxy.update_proxy()
                self.cfg['queue']['x509proxy'] = self.proxy.get_proxy()
            except Exception:
                logger.warning('cannot setup x509 proxy', exc_info=True)

        while True:
            start = time.monotonic()
            await asyncio.to_thread(blocking())  # type: ignore
            end = time.monotonic()
            await asyncio.sleep(max(1, 300 - (end-start)))



from datetime import datetime
import logging
from unittest.mock import AsyncMock, MagicMock
from uuid import uuid4

import pytest

from iceprod.common.mongo_queue import Message
from iceprod.core.config import Config
from iceprod.core.jsonUtil import json_encode
import iceprod.services.actions.submit
from iceprod.services.actions.submit import Action, get_scope


@pytest.mark.parametrize('config,error', [
    ({
        'tasks':[{
            'name': 'testing',
            'trays': [{
                'modules': [{
                    'src': '/usr/bin/python3',
                    'args': ''
                }]
            }]
        }],
        'version': 3.2,
    }, None),
    ({'tasks':[
        {
            'name': 'testing',
            'trays': []
        }
    ]}, 'Validation error'),
    ({
        'tasks':[{
            'name': 'testing',
            'trays': [{
                'modules': [{
                    'src': '/usr/bin/python3',
                    'args': '',
                    'data': []
                }]
            }]
        }],
        'version': 3.1,
    }, 'Validation error'),
])
async def test_submit_server_request(config, error, server):
    client = server(roles=['user'], groups=['users'])
    args = {
        'config': json_encode(config),
        'description': 'description',
        'jobs_submitted': 10,
    }
    if error:
        with pytest.raises(Exception, match=error):
            await client.request('POST', '/actions/submit', args)
    else:
        ret = await client.request('POST', '/actions/submit', args)
        assert 'result' in ret

        ret = await client.request('GET', f'/actions/submit/{ret["result"]}')
        assert ret['status'] == 'queued'


async def test_submit_server_request_bad_args(server):
    client = server(roles=['user'], groups=['users'])
    args = {}
    with pytest.raises(Exception):
        await client.request('POST', '/actions/submit', args)


@pytest.mark.parametrize('path,movement,scope', [
    ('/data/user/foo', 'input', 'storage.read:/data/user'),
    ('/data/user/foo', 'output', 'storage.modify:/data/user'),
    ('/data/user/foo', 'both', 'storage.modify:/data/user'),
    ('', 'input', 'storage.read:/'),
    ('/data/user/$(iter)/foo', 'input', 'storage.read:/data/user'),
    ('/data/user$/$(iter)/foo', 'input', 'storage.read:/data'),
    ('/data/user/foo/000000-000999/bar', 'input', 'storage.read:/data/user/foo'),
    ('/data/exp/IceCube/2025/filtered/PFFilt/0612/PFFilt_PhysicsFiltering_Run00141027_Subrun00000000_00000074.tar.bz', 'input', 'storage.read:/data/exp/IceCube/2025/filtered/PFFilt'),
    ('/data/exp/IceCube/2025/filtered/dev/off.7/0612/Run00141027_89/Offline_IC86.2025_data_Run00141027_Subrun00000000_00000073.i3.zst.sha512', 'input', 'storage.read:/data/exp/IceCube/2025/filtered/dev/off.7'),
])
def test_submit_scope(path, movement, scope):
    assert get_scope(path, movement) == scope


@pytest.mark.parametrize('config,scope', [
    ({
        'tasks':[{
            'name': 'testing',
            'trays': [{
                'modules': [{
                    'src': '/usr/bin/python3',
                    'args': ''
                }]
            }],
            'data': [
                {
                    'remote': 'token:///data/sim/IceCube/2025/file.i3.zst',
                    'movement': 'input'
                }
            ]
        }],
        'version': 3.2,
    }, 'storage.read:/data/sim/IceCube/2025'),
    ({
        'tasks':[{
            'name': 'testing',
            'trays': [{
                'modules': [{
                    'src': '/usr/bin/python3',
                    'args': ''
                }]
            }],
            'data': [
                {
                    'remote': 'token:///data/sim/IceCube/2025/file.i3.zst',
                    'movement': 'input'
                }
            ],
            'token_scopes': {
                'token://': 'storage.write:/foo/bar'
            }
        }],
        'version': 3.2,
    }, 'storage.read:/data/sim/IceCube/2025 storage.write:/foo/bar')
])
async def test_submit_run(config, scope, monkeypatch):
    monkeypatch.setattr(iceprod.services.actions.submit, 'TOKEN_PREFIXES', {
        'token://': 'https://iceprod.tokens',
    })

    queue = AsyncMock()
    api_client = AsyncMock()
    cred_client = AsyncMock()
    submit = Action(queue, logging.getLogger('action'), api_client, cred_client)

    description = 'Test dataset'

    d = Config(config)
    d.fill_defaults()
    d.validate()
    config_valid = d.config

    api_client.request.return_value = {'result': 'dataset'}
    cred_client.request.return_value = 'thetoken'

    await submit.run(Message(
        payload={
            'config': json_encode(config_valid),
            'description': description,
            'jobs_submitted': 10,
            'username': 'user',
            'group': 'users',
            'extra_submit_fields': '{"always_active":true}',
        },
        uuid=uuid4().hex,
        status='queued',
        priority=0,
        created_at=datetime.now(),
    ))

    api_calls = api_client.request.call_args_list
    assert len(api_calls) == 2
    assert api_calls[0][0][1] == '/datasets'
    assert api_calls[0][0][2] == {
        'description': description,
        'jobs_submitted': 10,
        'tasks_submitted': 10,
        'tasks_per_job': 1,
        'username': 'user',
        'group': 'users',
        'always_active': True,
    }
    assert api_calls[1][0][1] == '/config/dataset'

    cred_calls = cred_client.request.call_args_list
    assert len(cred_calls) == 2
    assert cred_calls[0][0][1] == '/create'
    assert cred_calls[0][0][2]['scope'] == scope
    assert cred_calls[1][0][1] == '/datasets/dataset/credentials'
